#ifndef POINT2_H
#define POINT2_H

#include "rigid_motion.h"
#include <math.h>

namespace rescue
{

template <class arithmetic>
class Point2
{

public:

    typedef rigid_motion<arithmetic> Motion;
    typedef arithmetic arithmetic_t;

    arithmetic x;
    arithmetic y;

    Point2() {}

    Point2(const arithmetic _x, const arithmetic _y)
    {
        x = _x;
        y = _y;
    }

    arithmetic distance(const Point2<arithmetic>& q)
    {
        return sqrt((x-q.x)*(x-q.x)+(y-q.y)*(y-q.y));
    }

    Point2& operator *= (const Motion& m)
    {
        arithmetic x_new = x*m.cos_phi - y*m.sin_phi + m.t_x;
        arithmetic y_new = x*m.sin_phi + y*m.cos_phi + m.t_y;
        x = x_new;
        y = y_new;
        return *this;
    }

    friend std::ostream& operator<<(std::ostream&, const Point2&);

    static void get_representation(
        const std::pair<Point2<arithmetic>, Point2<arithmetic> >& p,
        Motion& g_p,
        arithmetic& r_p);

    bool operator==(const Point2<arithmetic>&) const;

};

template <class arithmetic>
std::ostream& operator<<(std::ostream& o, const Point2<arithmetic>& p)
{
    return o<<"("<<p.x<<","<<p.y<<")";
}

template<class arithmetic>
void Point2<arithmetic>::get_representation(
    const std::pair<Point2<arithmetic>, Point2<arithmetic> >& p,
    Motion& g_p,
    arithmetic& r_p
   )
{
    arithmetic two = 2.0;

    Point2<arithmetic> v0 = p.first;
    Point2<arithmetic> v1 = p.second;

    // Compute a single group element G_<p0,p1>(<q0,q1>).
    
    arithmetic l_p = sqrt((v1.x-v0.x)*(v1.x-v0.x)+(v1.y-v0.y)*(v1.y-v0.y));
    arithmetic l_p_2 = l_p / two;
    r_p = l_p_2;

    Point2<arithmetic> c_p; // used as the line segment's centroid
    c_p.x = (v0.x+v1.x)/two;
    c_p.y = (v0.y+v1.y)/two;
    
    g_p.t_x = c_p.x;
    g_p.t_y = c_p.y;

    if (l_p>0.0)
    {
        arithmetic g_p_sin_phi = (v1.y-v0.y)/l_p;
        arithmetic g_p_cos_phi = (v1.x-v0.x)/l_p;
        arithmetic g_p_phi = atan(g_p_sin_phi / g_p_cos_phi);
        g_p *= Motion(g_p_sin_phi,g_p_cos_phi,0.0,0.0);
    } else {
        g_p.sin_phi = 0.0;
        g_p.cos_phi = 1.0;
    }

}



template<class arithmetic>
bool Point2<arithmetic>::operator==(const Point2<arithmetic>& p2) const
{
    return (x==p2.x && y==p2.y);
}



}

#endif
