
#define MAX_HITS 10

#include "CurveCanvas.h"
#include <qpainter.h>
#include <qmessagebox.h>
#include <iostream>
#include <ReSCue/Point2.h>
#include <ReSCue/polygonal_curve.h>
#include <ReSCue/covering_matcher.h>
#include <ReSCue/sampling_matcher.h>
#include <ReSCue/circular_arc_edge.h>
#include <ReSCue/free_space_graph.h>
#include <ReSCue/extended_fs_graph.h>
#include <ReSCue/partial_match_graph.h>
#include <ReSCue/partial_sampling_matcher.h>
#include <ReSCue/lcsc_match_graph.h>
#include <ReSCue/lcsc_sampling_matcher.h>
#include <ReSCue/directed_hausdorff_graph.h>

typedef Point2<double> Point;

//typedef covering_matcher<double,s1_interval_container<double> > CoveringMatcher;
typedef sampling_matcher<
            free_space_graph<
                circular_arc_edge<int,Point> 
            > 
        > 
        SamplingMatcher;
typedef SamplingMatcher::curve Curve;
typedef SamplingMatcher::vertex Vertex;
typedef SamplingMatcher::point Point;
typedef sampling_matcher <
            extended_fs_graph <
                circular_arc_edge<int,Point> 
            > 
        > 
        FastSamplingMatcher;

typedef partial_sampling_matcher <
            partial_match_graph <
                circular_arc_edge<int,Point> 
            > 
        > 
        PartialSamplingMatcher;

typedef lcsc_sampling_matcher <
            lcsc_match_graph <
                circular_arc_edge<int,Point> 
            > 
        > 
        LcscSamplingMatcher;


typedef partial_sampling_matcher <
            directed_hausdorff_graph <
                circular_arc_edge<int,Point> 
            > 
        > 
        HausdorffSamplingMatcher;


typedef Curve::vertex Vertex;
typedef Curve::vertex_iterator curve_vertex_iterator;


CurveCanvas::CurveCanvas( QWidget *parent, const char *name )
        : QWidget( parent, name )
{
    setPalette( QPalette( QColor( 250, 250, 200) ) );
    color_id = 0;
    eps = 10.0;
    eps_prime = 20.0;
    match_found = false;
	reparametrizations_available = false;

    matching_method = "fast sampling";
}


void CurveCanvas::paintEvent( QPaintEvent * )
{

    QPainter qp( this );
    qp.setPen( QColor(0,0,0) );
    qp.setBrush( QColor(255,0,0) );

    Curve::vertex_iterator v_it;
    Vertex* prev = NULL;

    qp.drawLine(10,10,10,20);
    qp.drawLine(10,15,10+eps,15);
    qp.drawLine(10+eps,10,10+eps,20);

    qp.drawLine(10,40,10,50);
    qp.drawLine(10,45,10+eps_prime,45);
    qp.drawLine(10+eps_prime,40,10+eps_prime,50);

    for (v_it = red_curve.vertex_begin(); v_it!=red_curve.vertex_end(); ++v_it)
    {
        Point p = (*v_it).pos;
        if (prev!=NULL)
        {
            Point q = prev->pos;
            qp.drawLine(q.x,q.y,p.x,p.y);
        }
        QRect myRect(p.x-3, p.y-3, 6, 6);
        qp.drawRect( myRect );          
        prev = &(*v_it);
    }

    qp.setBrush( QColor(0,0,255) );
    prev = NULL;

    for (v_it = blue_curve.vertex_begin(); v_it!=blue_curve.vertex_end(); ++v_it)
    {
        Point p = (*v_it).pos;
        if (prev!=NULL)
        {
            Point q = prev->pos;
            qp.drawLine(q.x,q.y,p.x,p.y);
        }
        QRect myRect(p.x-3, p.y-3, 6, 6);
        qp.drawRect( myRect );          
        prev = &(*v_it);
    }

    if (match_found) {
        qp.setPen( QColor(64,64,64) );
        qp.setBrush( QColor(128,128,128) );
        prev = NULL;

        for (v_it = result_curve.vertex_begin(); v_it!=result_curve.vertex_end(); ++v_it)
        {
            Point p = (*v_it).pos;
            if (prev!=NULL)
            {
                Point q = prev->pos;
                qp.drawLine(q.x,q.y,p.x,p.y);
            }
            QRect myRect(p.x-3, p.y-3, 6, 6);
            qp.drawRect( myRect );          
            prev = &(*v_it);
        }
    }

    if (reparametrizations_available)
    {
        qp.setPen( QColor(0,0,0) );
        qp.setBrush( QColor(0,255,0) );
        Point p = (*red_kappa_pos).pos;
		Point q = (*blue_lambda_pos).pos;
        QRect myRect1(p.x-3, p.y-3, 6, 6);
        QRect myRect2(q.x-3, q.y-3, 6, 6);
        qp.drawRect( myRect1 );          
        qp.drawRect( myRect2 );          
        qp.drawLine(q.x,q.y,p.x,p.y);
    }

}

void CurveCanvas::next_reparametrization()
{
    if (reparametrizations_available)
    {
		red_kappa_pos++;
		if (red_kappa_pos==red_kappa.vertex_end())
			red_kappa_pos = red_kappa.vertex_begin();
		blue_lambda_pos++;
		if (blue_lambda_pos==blue_lambda.vertex_end())
			blue_lambda_pos = blue_lambda.vertex_begin();
    }
	repaint();
}

void CurveCanvas::prev_reparametrization()
{
    if (reparametrizations_available)
    {
		if (red_kappa_pos==red_kappa.vertex_begin())
			return;	
		red_kappa_pos--;
		if (blue_lambda_pos==blue_lambda.vertex_begin())
			return;
		blue_lambda_pos--;
    }
	repaint();
}


QSizePolicy CurveCanvas::sizePolicy() const
{
    return QSizePolicy( QSizePolicy::Expanding, QSizePolicy::Expanding );
}

void CurveCanvas::mouseReleaseEvent( QMouseEvent *e )
{
    match_found = false;

    if ( e->button() != LeftButton )
        return;

    QPoint pnt = e->pos();
    double myX = pnt.x();
    double myY = pnt.y();

    Vertex* v = new Vertex(Point(myX,myY));

    if (color_id==0)    
        red_curve.push_back(v);
    else if (color_id==1)    
        blue_curve.push_back(v);

    repaint();
    
}
    
void CurveCanvas::setColor(int col_id)
{
    color_id = col_id;
    if (col_id==0)
        color = QColor(255,0,0);
    else if (col_id==1)
        color = QColor(0,0,255);

}

void CurveCanvas::del()
{
    match_found = false;
	reparametrizations_available = false;
    if (color_id==0)    
        red_curve.delete_last_point();
    else if (color_id==1)    
        blue_curve.delete_last_point();
    repaint();
}

void CurveCanvas::clear_curve()
{
    match_found = false;
	reparametrizations_available = false;
    if (color_id==0)    
        red_curve.clear();
    else if (color_id==1)    
        blue_curve.clear();
    repaint();
}

void CurveCanvas::motion(const Motion& m)
{

    if (color_id==0)    
        red_curve *= m;
    else
        blue_curve *= m;
    repaint();

}

void CurveCanvas::match_now(double eps, double eps_prime)
{

    if (matching_method=="fast sampling")
    {
        FastSamplingMatcher M(red_curve,blue_curve,eps,eps_prime);

        match_found = M.match();
        reparametrizations_available = M.get_reparametrized_curves(red_kappa,blue_lambda);

        if (match_found)
        {
			red_kappa_pos = red_kappa.vertex_begin();
			blue_lambda_pos = blue_lambda.vertex_begin();
            QMessageBox::information( this, "ReSCue", "A Match was found.");
            // and now: process result...
            result_curve = blue_curve;
            result_curve *= M.get_match();
            repaint();
        } else {
            QMessageBox::information( this, "ReSCue", "No Match found.");
        }

        M.clear();
    } else if (matching_method=="standard sampling") {
        SamplingMatcher M(red_curve,blue_curve,eps,eps_prime);

        match_found = M.match();

        if (match_found)
        {
            QMessageBox::information( this, "ReSCue", "A Match was found.");
            // and now: process result...
            result_curve = blue_curve;
            result_curve *= M.get_match();
            repaint();
        } else {
            QMessageBox::information( this, "ReSCue", "No Match found.");
        }

        M.clear();
    } else if (matching_method=="standard sampling discrete") {
        SamplingMatcher M(red_curve,blue_curve,eps);

        match_found = M.match();

        if (match_found)
        {
            QMessageBox::information( this, "ReSCue", "A Match was found.");
            // and now: process result...
            result_curve = blue_curve;
            result_curve *= M.get_match();
            repaint();
        } else {
            QMessageBox::information( this, "ReSCue", "No Match found.");
        }

        M.clear();
    } else if (matching_method=="exp-time sampling (be warned!)") {
        //CoveringMatcher M(red_curve,blue_curve,eps,eps_prime);

        //match_found = M.match();

        if (match_found)
        {
            QMessageBox::information( this, "ReSCue", "A Match was found.");
            // and now: process result...
            result_curve = blue_curve;
            //result_curve *= M.get_match();
            repaint();
        } else {
            QMessageBox::information( this, "ReSCue", "No Match found.");
        }

        //M.clear();
    } else if (matching_method=="partial match") {
        PartialSamplingMatcher M(red_curve,blue_curve,eps,eps_prime);

        match_found = M.match();
        reparametrizations_available = M.get_reparametrized_curves(red_kappa,blue_lambda);

        if (match_found)
        {
			red_kappa_pos = red_kappa.vertex_begin();
			blue_lambda_pos = blue_lambda.vertex_begin();
            QMessageBox::information( this, "ReSCue", "A Match was found.");
            // and now: process result...
            result_curve = blue_curve;
            result_curve *= M.get_match();
            repaint();
        } else {
            QMessageBox::information( this, "ReSCue", "No Match found.");
        }

        M.clear();
    } else if (matching_method=="directed hausdorff match") {
        HausdorffSamplingMatcher M(red_curve,blue_curve,eps);

        match_found = M.match();
        reparametrizations_available = M.get_reparametrized_curves(red_kappa,blue_lambda);

        if (match_found)
        {
			red_kappa_pos = red_kappa.vertex_begin();
			blue_lambda_pos = blue_lambda.vertex_begin();
            QMessageBox::information( this, "ReSCue", "A Match was found.");
            // and now: process result...
            result_curve = blue_curve;
            result_curve *= M.get_match();
            repaint();
        } else {
            QMessageBox::information( this, "ReSCue", "No Match found.");
        }

        M.clear();
    } else if (matching_method=="common subcurve match") {
        LcscSamplingMatcher M(red_curve,blue_curve,eps,eps_prime);
        match_found = M.match();
        reparametrizations_available = M.get_reparametrized_curves(red_kappa,blue_lambda);

        if (match_found)
        {
			red_kappa_pos = red_kappa.vertex_begin();
			blue_lambda_pos = blue_lambda.vertex_begin();
            QMessageBox::information( this, "ReSCue", "A Match was found.");
            // and now: process result...
            result_curve = blue_curve;
            result_curve *= M.get_match();
            repaint();
        } else {
            QMessageBox::information( this, "ReSCue", "No Match found.");
        }

        M.clear();
    } else {
        QMessageBox::information( this, "ReSCue", "Matching method not available.");
    }
}

void CurveCanvas::open(const QString& fname)
{
    
    std::fstream ifs(fname,std::ios::in | std::ios::binary);
    if (!ifs)
    {
        QMessageBox::information( this, "ReSCue", "File not found.\n");
        return;
    }
    if (color_id==0)
        red_curve.read(ifs);
    else
        blue_curve.read(ifs);
    repaint();

}

void CurveCanvas::save(const QString& fname)
{
    std::fstream ofs(fname,std::ios::out | std::ios::binary);
    if (!ofs)
    {
        QMessageBox::information( this, "ReSCue", "Could not open file.\n");
        return;
    }
    if (color_id==0)
        red_curve.write(ofs);
    else
        blue_curve.write(ofs);
}

void CurveCanvas::change_method(const QString& mthd)
{
    matching_method = mthd;
}

void CurveCanvas::setEps(double e)
{
    eps = e;
    repaint();
}

void CurveCanvas::setEps_prime(double e)
{
    eps_prime = e;
    repaint();
}
