
/****************************************************************
**
** Definition of CurveCanvas class, Qt tutorial 8
**
****************************************************************/

#ifndef RESCUE__CURVECANVAS_H
#define RESCUE__CURVECANVAS_H

#include <qwidget.h>
#include <ReSCue/polygonal_curve.h>
#include <ReSCue/Point2.h>

using namespace rescue;

typedef polygonal_curve<Point2<double> > Curve;
typedef Curve::motion Motion;

/** The \c CurveCanvas class provides a Qt-widget with the functionality for drawing two
polygonal curves, a red one and a blue one. Different slots enable deletion of points, 
clearing a curve, reading and writing curves from and to a file, as well as matching the two 
curves. The paramters for matching can be set by slots.
*/
class CurveCanvas : public QWidget
{
    Q_OBJECT
 
public:
    
    CurveCanvas( QWidget *parent=0, const char *name=0 );

    QSizePolicy sizePolicy() const;

public slots:
    
    void setColor(int);
    void del();
    void clear_curve();
    void next_reparametrization();
    void prev_reparametrization();
    void motion(const Motion&);
    void match_now(double,double);
    void open(const QString&);
    void save(const QString&);
    void setEps(double);
    void setEps_prime(double);
    void change_method(const QString&);

protected:
    
    void paintEvent( QPaintEvent * );
    void mouseReleaseEvent( QMouseEvent * );

private:
    
    int color_id;
    QColor color;

    QString matching_method;

    Curve red_curve;
    Curve blue_curve;

    /** \c result_curve is a copy of \c blue_curve, transformed by a matching group element --
    if such was found during the last call of \c match_now().
    */
    Curve result_curve;

    /** If the matching method used is capable of computing reparametrizations that indicate a
    match, then reparametrized curves \c red_kappa and \c blue_lambda (both of same lenght) 
    are computed, so that the corresponding points of the two curves are identified with each
    other by the reparametrizations found while matching.
    */
    Curve red_kappa;
    
    /** If the matching method used is capable of computing reparametrizations that indicate a
    match, then reparametrized curves \c red_kappa and \c blue_lambda (both of same lenght) 
    are computed, so that the corresponding points of the two curves are identified with each
    other by the reparametrizations found while matching.
    */
    Curve blue_lambda;

    double eps;
    double eps_prime;

    /** If the last call of of \c match_now() found a match between \c red_curve and 
    \c blue_curve, \c match_found is \c true. Otherwise (and initially), \c match_found 
    is \c false.
    */
    bool match_found;

    /** If the matching method used is capable of computing reparametrizations, so that
    \c red_kappa and \c blue_lambda are available, this value is set \c true, otherwise
    it is set \c false.
    */
    bool reparametrizations_available;

    /** Indicates the position of the current reparametrization pair on \c red_curve.
    */
	Curve::vertex_iterator red_kappa_pos;

    /** Indicates the position of the current reparametrization pair on \c blue_curve.
    */
	Curve::vertex_iterator blue_lambda_pos;

};

#endif // CURVECANVAS_H
