/****************************************************************
**
** Definition of CurveParameters class, Qt tutorial 8
**
****************************************************************/

#ifndef RESCUE__CURVEPARAMETERS_H
#define RESCUE__CURVEPARAMETERS_H

#include <qvbox.h>
#include <qwidget.h>
#include <ReSCue/Point2.h>
#include <ReSCue/polygonal_curve.h>
#include "EpsControl.h"

#include <ReSCue/sampling_matcher.h>
#include <ReSCue/circular_arc_edge.h>
#include <ReSCue/free_space_graph.h>
#include <ReSCue/extended_fs_graph.h>
#include <ReSCue/directed_hausdorff_graph.h>
#include <ReSCue/partial_match_graph.h>
#include <ReSCue/partial_sampling_matcher.h>


class QPushButton;
class QButtonGroup;

class QPushButton;

using namespace rescue;

typedef polygonal_curve<Point2<double> > Curve;
typedef Curve::motion Motion;

/** The \c CurveParameters class provides a Qt-widget for controlling the paramters of a
\c CurveCanvas.
*/
class CurveParameters : public QWidget
{

    Q_OBJECT

public:
    CurveParameters( QWidget *parent=0, const char *name=0 );

    int value() const;

public slots:

	void next_pair();
	void prev_pair();
    void del();
    void clear_curve();
    void move(int);
    void rotate_left();
    void rotate_right();
    void select_red();
    void select_blue();
    void match();
    void open();
    void testseries();
    void save();
    void setEps(double);
    void setEps_prime(double);
    void method_changed(const QString&);

signals:

	void show_next_pair();
	void show_prev_pair();
    void colorChanged( int );
    void remove_last_point();
    void clear_active_curve();
    void match_now(double,double);
    void rigid_motion(const Motion&);
    void open_file(const QString&);
    void save_file(const QString&);
    void start_testseries(const QString&);
    void epsChanged(double);
    void eps_primeChanged(double);
    void change_method(const QString&);

private:
    QPushButton  *btn1;
    QPushButton  *btn2;
    QButtonGroup *curve_select;
    QPushButton *button_red;
    QPushButton *button_blue;
    
    QPushButton *qsave;
    QPushButton *qopen;

    double eps;
    double eps_prime;

};

#endif // CurveParameters_H
