/****************************************************************
**
** Qt tutorial 8
**
****************************************************************/

#include <qapplication.h>
#include <qpushbutton.h>
#include <qlcdnumber.h>
#include <qfont.h>
#include <qlayout.h>

#include "CurveParameters.h"
#include "CurveCanvas.h"
#include "testseries.h"

class MyWidget: public QWidget
{
public:
    MyWidget( QWidget *parent=0, const char *name=0 );
};

MyWidget::MyWidget( QWidget *parent, const char *name )
        : QWidget( parent, name )
{

    CurveParameters* parameters = new CurveParameters( this, "parameters" );

    CurveCanvas *cc
        = new CurveCanvas( this, "CurveCanvas" );

    connect( parameters, SIGNAL(colorChanged(int)),
             cc, SLOT(setColor(int)) );

    connect( parameters, SIGNAL(change_method(const QString&)),
             cc, SLOT(change_method(const QString&)) );

    connect( parameters, SIGNAL(open_file(const QString&)),
             cc, SLOT(open(const QString&)) );

    connect( parameters, SIGNAL(save_file(const QString&)),
             cc, SLOT(save(const QString&)) );

    connect( parameters, SIGNAL(remove_last_point()),
             cc, SLOT(del()) );

    connect( parameters, SIGNAL(clear_active_curve()),
             cc, SLOT(clear_curve()) );

    connect( parameters, SIGNAL(match_now(double,double)),
             cc, SLOT(match_now(double,double)) );

    connect( parameters, SIGNAL(rigid_motion(const Motion&)),
             cc, SLOT(motion(const Motion&)) );

    connect( parameters, SIGNAL(epsChanged(double)),
             cc, SLOT(setEps(double)) );

    connect( parameters, SIGNAL(eps_primeChanged(double)),
             cc, SLOT(setEps_prime(double)) );

    connect( parameters, SIGNAL(show_next_pair()),
             cc, SLOT(next_reparametrization()) );

    connect( parameters, SIGNAL(show_prev_pair()),
             cc, SLOT(prev_reparametrization()) );

    QGridLayout *grid = new QGridLayout( this, 2, 2, 10 );
    //2x2, 10 pixel border

    grid->addWidget( parameters, 1, 0, Qt::AlignTop );
    grid->addWidget( cc, 1, 1 );
    grid->setColStretch( 1, 10 );

    parameters->setFocus();
}

int main_run( int argc, char **argv )
{

    if (argc>1)
    {
        // Batch-Mode -- no GUI start 
        for (int i=1; i<argc; i++)
        {
            std::cout<<"processing batch-file "<<argv[i]<<"...\n";
            my_testseries(argv[i]);
        }
        std::cout<<"finished.\n";
        return 0;
    }

    QApplication a( argc, argv );

    MyWidget w;
    w.setGeometry( 100, 100, 500, 355 );
    a.setMainWidget( &w );
    a.setName("ReSCue - REtrieval System for CUrvEs. Axel Mosig, Univ. Bonn, 2003");

    w.show();
    int res_exec = a.exec();
    return res_exec;

}

int main( int argc, char **argv )
{

    char* cc = new char[2];
    int res_exec = main_run(argc,argv);

    return res_exec;

}
