#define DEBUG 0
#define STATS 0

#define EFFECTIVE_EPSILON    0     // get the effective epsilon
#define BIT_GRID             1     // replace the AlignmentGrid w/ BitGrid
#define SIMPLE_MESH          0     // use a simple grid instead of a mesh 
#define GET_BUCKETS_FASTSQR  0     // it's actually much slower this way !
#define GET_BUCKETS_LINEARLY 1     // the linear algo of Drago
#define TRIVIAL              0     // get all n^2 pairs
#define STAMP_ANNULUS        0     // stamp the annulus on the buckets
#define WANT_RESULTS         0     // Generate set of aligned points

#include <time.h>
#include <math.h>
#include <stdio.h>
#include "match.h"
#include <fstream>


double Algorithm_One::DefaultEpsilon = 1.0;
double Algorithm_One::DefaultBeta = 1.0;

double M_PI = 3.1415926535897932384626433832795;

// this is what we divide epsilon by for the alignment grid
//double AlignGridFactor = 1.4142135623730950488016887242097 / Algorithm_One::DefaultBeta;
//MeshFactor = 2;




// it needs to be global in order to work but why !???
ostream& operator<< (ostream& out, SPoint &p)
{return out << "(" << p.x << ", " << p.y << ")";}

// it needs to be global in order to work but why !???
ostream& operator<< (ostream& out, Index &n)
{return out << "[" << n.x << " , " << n.y << "]";}

ostream& operator<< (ostream& out, BoundingBox &box)
{return out << "{ " << box.min << ", " << box.max << " }";}

ostream& operator<< (ostream& out, Pair &pair)
{return out << "< " << pair.p1 << ", " << pair.p2 << " >";}

ostream& operator<< (ostream& out, Transformation &t)
{return out << "{translate = " << t.translateV << ", rotate = " 
	    << t.theta << "}";}

ostream& operator<< (ostream& out, FormattedResult &r)
{return out << "RESULT: " << r.transform << ", " << (long)
   r.transformedModel;}





#if ( EFFECTIVE_EPSILON && ! BIT_GRID )
List<double> *tempEpsilons = NULL;
List<List<double>*> minEpsilons;

static
double GetMaxMinEpsilon(void)
{
  double maxDist = 0;

  ListIterator< List<double>* > epslistIterator( &minEpsilons );
  for (; !epslistIterator.IsDone(); epslistIterator.Next()) {
    ListIterator<double> epss(epslistIterator.Item());
    for (; ! epss.IsDone(); epss.Next()) {
      if ( maxDist < epss.Item() ) 
	maxDist = epss.Item();
    }
  }

  return maxDist;
}
#endif // EFFECTIVE_EPSILON && ! BIT_GRID

/*
static
void
DisplayIndexCenter( SPoint min, Index n )
{
  char name[100];
  static FILE* gridOutFile = NULL;

  if (gridOutFile == NULL) {
    sprintf( name, "gridout");
    gridOutFile = fopen( name, "w" );
  }
  assert( gridOutFile );

  SPoint center ( n.x * cellSizeOnX + cellSizeOnX / 2.0, 
		 n.y * cellSizeOnY + cellSizeOnY / 2.0 );
  center += min;
  fprintf( gridOutFile, "%lf %lf\n", center.x, center.y );
}


static
void
DisplayPoint( SPoint p )
{
  char name[100];
  static FILE* gridOutFile = NULL;

  if (gridOutFile == NULL) {
    sprintf( name, "gridout");
    gridOutFile = fopen( name, "w" );
  }
  assert( gridOutFile );

  fprintf( gridOutFile, "%lf %lf\n", p.x, p.y );
}

static
void
DisplayPair( Pair p )
{
  char name[100];
  static FILE* gridOutFile = NULL;

  if (gridOutFile == NULL) {
    sprintf( name, "gridout");
    gridOutFile = fopen( name, "w" );
  }
  assert( gridOutFile );

  fprintf( gridOutFile, "%lf %lf %lf %lf \n", p.p1.x, p.p1.y, p.p2.x, p.p2.y);
}
*/


void 
Transformation::SelfTest(void)
{
  cout << endl << "Begin Transformation SelfTest ... " << endl;

  cout << "Testing GetTransform/Transform. " << endl;
  SPoint p1 ( GetRandomValue(-500, 100), GetRandomValue(-500, 100) );
  SPoint p2 ( GetRandomValue(p1.x + 100, 1000), 
	     GetRandomValue(p1.y + 100, 1000) );
  
  SPoint translate ( GetRandomValue(-500, 500), GetRandomValue(-500, 500) );
  double rotate =  GetRandomValue(-2*M_PI, 2*M_PI);
  Transformation randomT ( translate, rotate );
  cout << "The random transformation was " << randomT << endl;
 
  Pair p( p1, p2);
  Pair q = randomT.Transform( p );

  cout << "To get from " << p << " to " << q << endl;

  GetTransform( p , q ); // now we have the transform saved in *this
  cout << "  we need the transform " << *this << endl;

  Pair result = Transform( p );
  cout << "So after we apply it to the 'from' pair, we get " << result << endl;

  Pair error = result - q;
  cout << "The error was " << error << endl;

  cout << "End Transformation SelfTest." << endl;
}

inline Transformation
Transformation::Compose( Transformation &other )
{
  Transformation trans ( this->translateV + other.translateV,
			 this->theta + other.theta );
  return trans;
}


inline SPoint
Transformation::Transform( SPoint p ) const
{
  SPoint result;
  SPoint translation = translateV;

  result = Rotate( p, theta );
  result += translation;

  return result;
}

inline Pair
Transformation::Transform( Pair pair ) const
{
  Pair result ( Transform( pair.p1 ), Transform ( pair.p2 ) );
  return result;
}


List<SPoint> *
Transformation::Transform( List<SPoint> *list ) const
{
  List<SPoint> *newlist = new List<SPoint>;
  assert( list );
  ListIterator<SPoint> iter( list ); 
  
  for (; !iter.IsDone(); iter.Next()) {
    SPoint pt = iter.Item();
    pt = this->Transform( pt );
    newlist->Append( pt );
  }
  return newlist;
}


// -------------------
// Transformation::GetTransform
// -------------------
// The public access method to get a transformation that will transform
// modelPair into imagePair.
// We rotate p1,p2 so that it's parallel to q1,q2 and then translate 
// it. This transformation assumes rotation about (0,0), so 
// everything works out..
// Do the same for p1,p2 and q2,q1 (to get the other possible matching)
void
Transformation::GetTransform( Pair modelPair, Pair imagePair)
{
  // To align the model pair (p1,p2) to the image pair (q1,q2), ...

  double alpha_model = GetAngleWithXaxis(modelPair);
  double alpha_image = GetAngleWithXaxis(imagePair);

  // Having done this for alphaP, alphaQ, the procedure for rotation is:
  // 1. Compute Delta = alphaQ - alphaP (because we transform *model* points)
  // 2. Apply a rotation of Delta to p1,p2, yielding p1', p2'
  // 3. Compute t = q1-p1'

  double delta = alpha_image - alpha_model;
  Pair modelPair_rotated = Rotate( modelPair, delta );
  SPoint t = imagePair.p1 - modelPair_rotated.p1; 
  
  // The transformation that we obtain is (t,Delta).
  this->translateV = t;
  this->theta = delta;
}


double 
Transformation::GetAngleWithXaxis(Pair pair) const
{
 
  // Let Vp be the vector p2-p1, and Vq be the vector q2-q1.
  // Let Vp = (xp,yp), Vq = (xq,yq). 

  SPoint p = pair.p2 - pair.p1;

  // Let 
  // alpha_p = arccos xp/sqrt(xp^2+yp^2)
  // alpha_q = arccos xq/sqrt(xq^2+yq^2)
  // Note that alpha_{p,q} represent the angles that the vectors make 
  // with the x axis s.t. 0 < alpha_{p,q} < PI/2.

  double alpha_p = acos( fabs(p.x) / sqrt( SQR(p.x) + SQR(p.y) ) );

  // The way to determine the actual signs of alpha_{p,q} is as follows:
  // We use the convention that angles measured counterclockwise from the x
  // axis are positive.

  // Now, for alpha_{p,q}:
  // 1) if xp,yp > 0, then 0 < alpha_p < PI/2
  // 2) If xp < 0, yp > 0, then PI/2 < alpha_p < PI
  // 3) If xp,yp < 0, then PI < alpha_p < 3 PI/2
  // 4) if xp > 0, yp < 0, then 3 PI/2 < alpha_p < 2 PI.

  // Now: from the cases above
  // 1) alphaP = aP
  // 2) alphaP = PI-aP
  // 3) alphaP = PI+aP
  // 4) alphaP = 2 PI - aP

  // ??? I have introduced the convention to bunch up '=' with '>' than 0
  // ??? is this OK ???

  if ( p.x >= 0 && p.y >= 0 ) ; // alpha_p = alpha_p
  else if ( p.x < 0 && p.y >= 0 ) alpha_p = M_PI - alpha_p;
  else if ( p.x < 0 && p.y < 0 ) alpha_p = M_PI + alpha_p;
  else if ( p.x >= 0 && p.y < 0 ) alpha_p = 2*M_PI - alpha_p;
  else assert( !"We missed a case in the angle direction determination" );

  return alpha_p;
}

inline BoundingBox
Transformation::Rotate( BoundingBox box, double angle )
{
  BoundingBox rotatedBox ( Rotate( box.min, angle ), 
			   Rotate( box.max, angle ) );
  return rotatedBox;
}

inline Pair 
Transformation::Rotate( Pair pair, double angle )
{
  Pair rotatedPair ( Rotate( pair.p1, angle ), 
		     Rotate( pair.p2, angle ) );
  return rotatedPair;
}

inline SPoint 
Transformation::Rotate( SPoint p, double angle )
{
  if ( angle == 0.0 ) 
    return p;

  SPoint newPt;

  newPt.x = cos(angle) * p.x - sin(angle) * p.y;
  newPt.y = sin(angle) * p.x + cos(angle) * p.y;

  return newPt;
}

// Get a point from the reference frame of the point p at distance radius
// and angle of rotation theta then retranslate in the common coordinate
// system
SPoint
Transformation::GetPoint(SPoint center, double radius, double angle)
{
  SPoint p;
  p.x = center.x + radius * cos(angle);
  p.y = center.y + radius * sin(angle);

  return p;
}

inline 
double
Transformation::LengthSQR( SPoint p1, SPoint p2 )
{
  return SQR(p1.x - p2.x) + SQR(p1.y - p2.y);
}

inline 
bool
Transformation::IsLengthBetweenSQR( double length,
				    double from, double to )
{
  // NOTE: all are SQUARED parameters
  // this is done to speed up the processing

  return ( length >= from  && length <=  to );
}

inline bool
Transformation::IsLengthBetweenSQR(SPoint p1, SPoint p2, 
				   double from, double to)
{
  // NOTE: from and to are SQUARED parameters
  // this is done to speed up the long parameterized checks where 'from' and
  // 'too' don't change in the loop calling this and they can be pre-squared

  double squaredDistance = SQR(p1.x - p2.x) + SQR(p1.y - p2.y);
  return IsLengthBetweenSQR( squaredDistance, from, to );
}

inline bool
Transformation::IsLengthBetweenSQR(Pair pair, double from, double to)
{
  return IsLengthBetweenSQR( pair.p1, pair.p2, from, to );
}


void
BoundingBox::MakeIntoSquare( void )
{
#if DEBUG
  //cout << "MakeIntoSquare: the box before: " << *this << endl;
#endif
  // pick the index of the sides with bigger distance in between
  double lengthX = max.x - min.x;
  double lengthY = max.y - min.y;
  double deltaLength = fabs(lengthX - lengthY);

  if (deltaLength == 0.0) return; // it's already a square
  if (lengthX > lengthY) {
    min.y -= deltaLength/2.0;
    max.y += deltaLength/2.0;
  } else {
    min.x -= deltaLength/2.0;
    max.x += deltaLength/2.0;
  }
#if DEBUG
  //cout << "MakeIntoSquare: the box after: " << *this << endl;
#endif
}

void
BoundingBox::EnlargeByFactor( double factor )
{
#if DEBUG
  //cout << "EnlargeByFactor: the box before: " << *this << endl;
#endif
  double lengthX = max.x - min.x;
  double lengthY = max.y - min.y;

  double newLengthX = lengthX * factor;
  double newLengthY = lengthY * factor;

  double deltaX = newLengthX - lengthX;
  double deltaY = newLengthY - lengthY;

  min.x -= deltaX/2.0;
  max.x += deltaX/2.0;

  min.y -= deltaY/2.0;
  max.y += deltaY/2.0;
#if DEBUG
  //cout << "EnlargeByFactor: the box after: " << *this << endl;
#endif
}

void
BoundingBox::EnlargeWith( double delta )
{
#if DEBUG
  //cout << "EnlargeWith: the box before: " << *this << endl;
#endif
  min.x -= delta;
  max.x += delta;

  min.y -= delta;
  max.y += delta;
#if DEBUG
  //cout << "EnlargeWith: the box after: " << *this << endl;
#endif
}


SPoint 
BoundingBox::GetCenter( void ) const
{
#if DEBUG
  // cout << "GetCenter: the box before: " << *this << endl;
#endif

  double lengthX = max.x - min.x;
  double lengthY = max.y - min.y;

  SPoint p( lengthX / 2.0 + min.x , lengthY / 2.0 + min.y);
  return p;
}


SPoint 
BoundingBox::SetCenter( double x, double y )
{
  SPoint newCenter( x, y );
  SPoint currentCenter = GetCenter();
  SPoint delta = newCenter - currentCenter;
#if DEBUG
  //cout << "SetCenter: Center of bounding box was " << currentCenter; 
#endif

  min += delta;
  max += delta;

  currentCenter = GetCenter();
#if DEBUG
  //cout << ", and is now set to " << currentCenter << endl; 
  //cout << "SetCenter: this is the bounding box after " << *this << endl;
#endif
  
  //assert( currentCenter == newCenter ); // the rounding errors can be brutal
  return currentCenter;
}


Grid::Grid(double _angle, BoundingBox _box, 
	   double _cellSizeOnX, double _cellSizeOnY)
{
  // which rotational step this grid covers
  angle = _angle;

  ptBox = _box;

  cellSizeOnX = _cellSizeOnX;
  cellSizeOnY = _cellSizeOnY;

  // maybe we don't need these values as class vars if we can 
  // always compute them from the box 
  areaSizeOnX = ptBox.max.x - ptBox.min.x;
  areaSizeOnY = ptBox.max.y - ptBox.min.y;

  // we take the floor of +1 and not ceil because if the division 
  // results in an integer, then we are in trouble -- a point lying
  // on the boundary will get assigned an index outside of the grid
  elementsOnX = floor( ( areaSizeOnX / cellSizeOnX ) + 1 );
  elementsOnY = floor( ( areaSizeOnY / cellSizeOnY ) + 1 );

#if DEBUG
  cout << "\tThe bounding box is " << ptBox << endl;
  cout << "\tThe cell size on X and Y are respectively " << cellSizeOnX <<
    ", " << cellSizeOnY << endl;
  cout << "\tThe area size on X and Y are respectively " << areaSizeOnX <<
    ", " << areaSizeOnY << endl;
  cout << "\tAllocating a grid with size " << elementsOnX
       << " by " << elementsOnY <<  " ..." << endl;
#endif

#if ( STATS && SIMPLE_MESH && ! TRIVIAL )
  cout << elementsOnX << " ";
#endif

  // allocate the 2-D array of dynamically allocated Lists
  grid = new List<SPoint>** [elementsOnX];
  for (int i = 0; i < elementsOnX; i++) {
    grid[i] = new List<SPoint>* [elementsOnY];
    for (int j = 0; j < elementsOnY; j++) {
      grid[i][j] = new List<SPoint>();
    }
  }

#if DEBUG 
  //  cout << "Grid #"<< id << " constructor: allocated " << 
  //  elementsOnX << " by " << elementsOnY << " cells." << endl;
#endif
}

Grid::~Grid()
{
  for (int i = 0; i < elementsOnX; i++) {
    for (int j = 0; j < elementsOnY; j++) {
      delete grid[i][j];
    }
    delete grid[i];
  }
  delete grid;
}

void
Grid::InitializeWith( List<SPoint> *points )
{
  assert( points );
  ListIterator<SPoint> pointIterator( points ); 
  for (; !pointIterator.IsDone(); pointIterator.Next()) {
    SPoint thePoint = pointIterator.Item();
    InsertPoint( thePoint );
#if DEBUG 
    //      cout << "Inserting the point " << thePoint << 
    //" in the alignment grid at angle " << angle << endl;
#endif      
  }

}

inline Index 
Grid::GetIndex(SPoint p) const
{
#if ( ! STAMP_ANNULUS )
  assert( PointIsOK(p) );
#endif
  SPoint adjustedPt = p - ptBox.min;
  Index n( floor( adjustedPt.x / cellSizeOnX ), 
	   floor( adjustedPt.y / cellSizeOnY ) );
#if DEBUG
  //cout << "SPoint " << p << " fits in bucket " << n << endl;
#endif
  return n;
}
  
void
Grid::InsertPoint(SPoint p)
{
  SPoint rotatedPt = p;

  // rotate to adjust for the position of the grid
  if ( angle != 0.0 )
    rotatedPt = Transformation::Rotate( p, -angle );

  // and calculate the index based on that
  Index n = GetIndex( rotatedPt );
  assert( IndexIsOK( n ) );

  // but still enter in the bucket the original point
  grid[n.x][n.y]->Append( p );
 
#if DEBUG
  //cout << "Inserting point " << p << " (after rotation " << rotatedPt
  //     << " ) in grid at angle " << angle << " at coordinates" << n << endl; 
#endif
}

/*
void 
DisplayIndexCenter( SPoint min, Index n, int quadrant, int taken )
{
  char name[100];
  assert( quadrant >= 1 && quadrant <= 4 && (taken == 0 || taken == 1) );
  if (gridOutFile[quadrant][taken] == NULL) {
    sprintf( name, "gridout%d%s", quadrant, ((taken)? "y": "n") );
    gridOutFile[quadrant][taken] = fopen( name, "w" );
  }
  assert( gridOutFile[quadrant][taken] );

  SPoint center ( n.x * cellSizeOnX + cellSizeOnX / 2.0, 
		 n.y * cellSizeOnY + cellSizeOnY / 2.0 );
  center += min;
  fprintf( gridOutFile[quadrant][taken], "%lf %lf\n", center.x, center.y );
}
*/


inline bool
Grid::PointIsOK(SPoint p) const
{
  return (p.x >= ptBox.min.x && p.x <= ptBox.max.x && 
	  p.y >= ptBox.min.y && p.y <= ptBox.max.y);
}

inline bool
Grid::IndexIsOK(Index n) const
{
  return (n.x >= 0 && n.x < elementsOnX &&
	  n.y >= 0 && n.y < elementsOnY);
}


SPoint 
Grid::GetPositionWithinBucket(SPoint p) const
{
  assert( PointIsOK(p) );
  SPoint adjustedPt = p - ptBox.min;
  SPoint n ((adjustedPt.x / cellSizeOnX) , (adjustedPt.y / cellSizeOnY));
  SPoint inside ((n.x - floor(n.x)), (n.y - floor(n.y)));
  assert( inside.x >= 0.0 && inside.x < 1.0 && 
	  inside.y >= 0.0 && inside.y < 1.0 );
  return inside;
}


List<SPoint>* 
Grid::GetBucket(SPoint samplePt) const
{
  if ( ! PointIsOK(samplePt) ) // return NULL if such point doesn't exist
    return NULL;
  Index n = GetIndex(samplePt); 
  return grid[n.x][n.y];
}



void
Grid::Get4Buckets(SPoint samplePt, List< List<SPoint>* > *buckets) const
{
  // rotate to adjust for the position of the grid
  samplePt = Transformation::Rotate( samplePt, -angle );

  if ( ! PointIsOK(samplePt) ) // return if such point doesn't exist
    return; 

  Index deltaIndex(0,0);
  SPoint inside = GetPositionWithinBucket(samplePt); 

  deltaIndex.x = ( (inside.x > 0.5)? 1: -1 );
  deltaIndex.y = ( (inside.y > 0.5)? 1: -1 );

  Index n = GetIndex(samplePt);
  Index n2 ( n.x + deltaIndex.x, n.y + deltaIndex.y );
  Index n3 ( n.x + deltaIndex.x, n.y );
  Index n4 ( n.x, n.y + deltaIndex.y );

#if DEBUG
  //  cout << "SPoint " << samplePt << " is within bucket " << n << 
  //  " at position " << inside << endl;
  //  cout << "Sampling indices ";
#endif

  assert( IndexIsOK( n ) ); // the point needs to have a bucket
  if ( ! grid[n.x][n.y]->IsEmpty() ) {
    buckets->Append( grid[n.x][n.y] );
    //cout << n << " ";
  }

  if ( IndexIsOK( n2 ) && ! grid[n2.x][n2.y]->IsEmpty() ) {
    buckets->Append( grid[n2.x][n2.y] );
    //cout << n2 << " ";
  }

  if ( IndexIsOK( n3 ) && ! grid[n3.x][n3.y]->IsEmpty() ) {
    buckets->Append( grid[n3.x][n3.y] );
    //cout << n3 << " ";
  }


  if ( IndexIsOK( n4 ) && ! grid[n4.x][n4.y]->IsEmpty() ) {
    buckets->Append( grid[n4.x][n4.y] );
    //cout << n4 << " ";
  }
  //cout << endl;
}

void
Grid::GetFBuckets(SPoint samplePt, List< List<SPoint>* > *buckets) const
{
  // rotate to adjust for the position of the grid
  samplePt = Transformation::Rotate( samplePt, -angle );

  if ( ! PointIsOK(samplePt) ) // return if such point doesn't exist
    return; 

  Index s = GetIndex(samplePt);
  for( int x = s.x - AlignGridFactor; x <= s.x + AlignGridFactor; x++ ) {
      
      for( int y = s.y - AlignGridFactor; y <= s.y + AlignGridFactor; y++ ) {
          Index n(x,y);

          // the index could be out of bounds (points on the periphery)
          if ( IndexIsOK( n ) && ! grid[ x ][ y ]->IsEmpty() ) {
              buckets->Append( grid[ x ][ y ] );
          }
      }
      
  }
}

// ==================================================================
#if ( ! STAMP_ANNULUS && SIMPLE_MESH )
// ==================================================================

// this fn keeps the points from going outside the grid
inline
SPoint
Grid::PointInGrid( SPoint p ) const
{
  SPoint newPt;

  if ( p.x >= ptBox.min.x && p.x <= ptBox.max.x )
    newPt.x = p.x;

  else if ( p.x < ptBox.min.x )
    newPt.x = ptBox.min.x;

  else if ( p.x > ptBox.max.x )
    newPt.x = ptBox.max.x;
  
  else assert( !"Something is wrong here" );


  if ( p.y >= ptBox.min.y && p.y <= ptBox.max.y )
    newPt.y = p.y;

  else if ( p.y < ptBox.min.y )
    newPt.y = ptBox.min.y;

  else if ( p.y > ptBox.max.y )
    newPt.y = ptBox.max.y;
  
  else assert( !"Something is wrong here" );

  return newPt;
}


void 
Grid::GetBucketsInAnulusFast( SPoint centerPoint, 
			      double radius, double epsilon,
			      List< List<SPoint>* > *allBuckets ) const
{
  int i, j;
  static SPoint outerRingVector ( epsilon + radius , epsilon + radius );

  // get the outer bounding box of the annulus
  SPoint outerFromPt = PointInGrid( centerPoint - outerRingVector );
  SPoint outerToPt = PointInGrid( centerPoint + outerRingVector );

  // get the inner bounding box of the annulus
  SPoint innerFromPt = PointInGrid( 
	       Transformation::GetPoint( centerPoint, 
					 radius - epsilon, 5*M_PI/4 ) );
  SPoint innerToPt = PointInGrid(
	       Transformation::GetPoint( centerPoint, 
					 radius - epsilon, M_PI/4 ) );

  // these points should be all valid since the bounding box was adjusted 
  // for the +- epsilon queries 

  Index innerFromIndex = GetIndex( innerFromPt );
  Index innerToIndex = GetIndex( innerToPt );
  Index pIndex = GetIndex( centerPoint );
  Index outerFromIndex = GetIndex( outerFromPt );
  Index outerToIndex = GetIndex( outerToPt );

  // Since we are going to go with a loop through all of these buckets and
  // test whether or not they cover the anulus, we don't need to assert
  // anything here, each bucket that falls between these boundaries will
  // be systematically evaluated whether or not it should be taken --
  // there should not be a chance for repeats

  // just handle the most obvious of the special cases
  if ( pIndex == outerFromIndex && pIndex == outerToIndex ) {
    allBuckets->Append( grid [ pIndex.x ] [ pIndex.y ] );
    return; // nothing else to do here
  }
  

  // the general strategy here is the following:
  // -------------------------------------------
  // don't worry about outlier cells, get all of them that are between the
  // outer annulus box and the inner anulus box
  // ---------------------------------------------

  // begin the traversal from the lower outside corner
  for( i = outerFromIndex.x; i <= innerFromIndex.x; i++ ) {
    for( j = outerFromIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );
      assert( IndexIsOK( n ) ); 

      allBuckets->Append( grid[ n.x ][ n.y ] );
    }
  }
  
  // from the current position
  for( ; i < innerToIndex.x; i++ ) {
    // we need to take care of two regions, so we have two for-loops
    for( j = outerFromIndex.y; j <= innerFromIndex.y; j++ ) {
      Index n ( i,j );
      assert( IndexIsOK( n ) ); 
      allBuckets->Append( grid[ n.x ][ n.y ] );
    }
    
    for( j = innerToIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );
      assert( IndexIsOK( n ) ); 
      allBuckets->Append( grid[ n.x ][ n.y ] );
    }
  }
  
  // from the current position
  for( ; i <= outerToIndex.x; i++ ) {
    for( int j = outerFromIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );
      assert( IndexIsOK( n ) ); 
      allBuckets->Append( grid[ n.x ][ n.y ] );
    }
  }
}

void 
Grid::GetBucketsInAnulus(SPoint centerPoint, double radius, double epsilon,
			 List< List<SPoint>* > *allBuckets ) const
{
  int i, j;
  SPoint closestComparePt, farthestComparePt, adjustedCenterPt;
  static SPoint outerRingVector ( epsilon + radius , epsilon + radius );


  // get the outer bounding box of the annulus
  SPoint outerFromPt = PointInGrid( centerPoint - outerRingVector );
  SPoint outerToPt = PointInGrid( centerPoint + outerRingVector );

  // get the inner bounding box of the annulus
  SPoint innerFromPt = PointInGrid( 
	       Transformation::GetPoint( centerPoint, 
					 radius - epsilon, 5*M_PI/4 ) );
  SPoint innerToPt = PointInGrid(
	       Transformation::GetPoint( centerPoint, 
					 radius - epsilon, M_PI/4 ) );

  // these points should be all valid since the bounding box was adjusted 
  // for the +- epsilon queries 

  Index innerFromIndex = GetIndex( innerFromPt );
  Index innerToIndex = GetIndex( innerToPt );
  Index pIndex = GetIndex( centerPoint );
  Index outerFromIndex = GetIndex( outerFromPt );
  Index outerToIndex = GetIndex( outerToPt );

  // Since we are going to go with a loop through all of these buckets and
  // test whether or not they cover the anulus, we don't need to assert
  // anything here, each bucket that falls between these boundaries will
  // be systematically evaluated whether or not it should be taken --
  // there should not be a chance for repeats

  // just handle the most obvious of the special cases
  if ( pIndex == outerFromIndex && pIndex == outerToIndex ) {
    allBuckets->Append( grid [ pIndex.x ] [ pIndex.y ] );
    return; // nothing else to do here
  }
  
  // begin the traversal from the lower outside corner
  for( i = outerFromIndex.x; i <= innerFromIndex.x; i++ ) {
    for( j = outerFromIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );

      if ( IsBucketInAnnulus( n, centerPoint, radius, epsilon ) ) {
	allBuckets->Append( grid[ n.x ][ n.y ] );
      }
    }
  }
    
  // from the current position
  for( ; i < innerToIndex.x; i++ ) {
    // we need to take care of two regions, so we have two for-loops
    for( j = outerFromIndex.y; j <= innerFromIndex.y; j++ ) {
      Index n ( i,j );

      if ( IsBucketInAnnulus( n, centerPoint, radius, epsilon ) ) {
	allBuckets->Append( grid[ n.x ][ n.y ] );
      }
    }

    for( j = innerToIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );

      if ( IsBucketInAnnulus( n, centerPoint, radius, epsilon ) ) {
	allBuckets->Append( grid[ n.x ][ n.y ] );
      }
   }
  }

  // from the current position
  for( ; i <= outerToIndex.x; i++ ) {
    for( int j = outerFromIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );

      if ( IsBucketInAnnulus( n, centerPoint, radius, epsilon ) ) {
	allBuckets->Append( grid[ n.x ][ n.y ] );
      }
    }
  }
}


// We are indebted to Drago for the algorithm behind this function
void 
Grid::GetBucketsInAnulusLinearly (SPoint centerPt, double radius, 
				  double epsilon,
				  List< List<SPoint>* > *allBuckets ) const
{
  double start;
  SPoint upperBeginPt ( centerPt.x, centerPt.y + (radius+epsilon) );
  SPoint lowerBeginPt ( centerPt.x, centerPt.y - (radius+epsilon) );
  SPoint rightEndPt ( centerPt.x + (radius+epsilon), centerPt.y );
  SPoint leftEndPt ( centerPt.x - (radius+epsilon), centerPt.y );

  // make sure we are not going outside the grid when we add the radius
  // there will be some incorrectness in the case when the points get out 
  // of the grid (when later we take their indices, we'll get some
  // indices that otherwise would not be there) but this seems OK
  // nevertheless since all buckets are tested for being within the
  // annulus

  upperBeginPt = PointInGrid( upperBeginPt );
  lowerBeginPt = PointInGrid( lowerBeginPt );
  rightEndPt = PointInGrid( rightEndPt );
  leftEndPt = PointInGrid( leftEndPt );

  // these points should be all valid since the bounding box was adjusted 
  // for the +- epsilon queries 

  // get the indices of these points
  Index upperBeginIndex = GetIndex( upperBeginPt );
  Index lowerBeginIndex = GetIndex( lowerBeginPt );
  Index rightEndIndex = GetIndex( rightEndPt );
  Index leftEndIndex = GetIndex( leftEndPt );


  // first we go through the 1st quadrant just to collect points
  // which we'll flip later to cover the other 3 quadrants of the annulus
  // and only then, we'll collect the buckets
  // ...
  // yeah, this would have been a good idea if the centerPoint was
  // always in the middle of a cell, which is NOT the case
  // so we need to just do it quadrant by quadrant


  // Quadrant 1
  //cout << "Quadrant 1" << endl;
  start = upperBeginIndex.y ; 

  for( int xDir = upperBeginIndex.x ; xDir <= rightEndIndex.x; xDir++ ) {
    int yDir = start; 
    assert( yDir >= rightEndIndex.y );
    
    // skip the possibly outlying initial buckets
    while( yDir >= rightEndIndex.y ) {
      Index n ( xDir, yDir );

      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	start = yDir; // next time start from here
	break; // just advance the index until it covers the annulus
      } 

      // handle the annulus going out of the grid
      else if ( n.x == upperBeginIndex.x ) {
	for( ; xDir < rightEndIndex.x &&
	       ! IsBucketInAnnulus( n, centerPt, radius, epsilon );
	     xDir ++, n.x ++ )
	  ; // end of for loop do nothing, just go until you find annulus 
	break;
      }

      yDir -- ;
    }
    
    // collect the buckets in the annulus
    while ( yDir >= rightEndIndex.y ) {
      Index n ( xDir, yDir );
  
      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	allBuckets->Append( grid[ n.x ][ n.y ] );
      }
      else
	break;
      
      yDir --;
    }
  }

  // Quadrant 2
  //cout << "Quadrant 2" << endl;

  start = upperBeginIndex.y ; 

  for( xDir = upperBeginIndex.x - 1 ; xDir >= leftEndIndex.x; xDir -- ) {
    int yDir = start; 
    assert( yDir >= leftEndIndex.y );

    // skip the possibly outlying initial buckets
    while( yDir >= rightEndIndex.y ) {
      Index n ( xDir, yDir );

      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	start = yDir; // next time start from here
	break; // just advance the index until it covers the annulus
      }      

      // handle the annulus going out of the grid
      else if ( n.x == upperBeginIndex.x ) {
	for( ; xDir > leftEndIndex.x &&
	       ! IsBucketInAnnulus( n, centerPt, radius, epsilon );
	     xDir -- , n.x -- )
	  ; // end of for loop. do nothing, just go until annulus is found
	break;
      }

      yDir -- ;
    }

    // collect the buckets in the annulus
    while ( yDir >= leftEndIndex.y ) {
      Index n ( xDir, yDir );
     
      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	allBuckets->Append( grid[ n.x ][ n.y ] );
      }
      else
	break;
      
      yDir --;
    }
  }


  // Quadrant 3
  //cout << "Quadrant 3" << endl;

  start = lowerBeginIndex.y ; 

  for( xDir = lowerBeginIndex.x ; xDir >= leftEndIndex.x; xDir -- ) {
    int yDir = start; 
    assert( yDir <= leftEndIndex.y );
    
    // skip the possibly outlying initial buckets
    while( yDir <= leftEndIndex.y ) {
      Index n ( xDir, yDir );

      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	start = yDir; // next time start from here
	break; // just advance the index until it covers the annulus
      }      

      // handle the annulus going out of the grid
      else if ( n.x == lowerBeginIndex.x ) {
	for( ; xDir > leftEndIndex.x &&
	       ! IsBucketInAnnulus( n, centerPt, radius, epsilon );
	     xDir -- , n.x -- )
	  ; // end of for loop. do nothing, just go until annulus is found
	break;
      }

      yDir ++ ;
    }

    // collect the buckets in the annulus
    while ( yDir < leftEndIndex.y ) {
      Index n ( xDir, yDir );
     
      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	allBuckets->Append( grid[ n.x ][ n.y ] );
      }
      else
	break;
      
      yDir ++ ;
    }
  }


  // Quadrant 4
  //  cout << "Quadrant 4" << endl;

  start = lowerBeginIndex.y ; 

  for(  xDir = lowerBeginIndex.x + 1 ; xDir <= rightEndIndex.x; xDir ++ ) {
    int yDir = start; 
    assert( yDir <= leftEndIndex.y );

   // skip the possibly outlying initial buckets
    while( yDir <= leftEndIndex.y ) {
      Index n ( xDir, yDir );

      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	start = yDir; // next time start from here
	break; // just advance the index until it covers the annulus
      }      

      // handle the annulus going out of the grid
      else if ( n.x == lowerBeginIndex.x ) {
	for( ; xDir < rightEndIndex.x &&
	       ! IsBucketInAnnulus( n, centerPt, radius, epsilon );
	     xDir ++, n.x ++ )
	  ; // end of for loop do nothing, just go until you find annulus 
	break;
      }

      yDir ++ ;
    }

    // collect the buckets in the annulus
    while ( yDir < leftEndIndex.y ) {
      Index n ( xDir, yDir );
     
      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	allBuckets->Append( grid[ n.x ][ n.y ] );
      }
      else
	break;
      
      yDir ++ ;
    }
  }

}



// ==================================================================
#elif ( SIMPLE_MESH && STAMP_ANNULUS )
// ==================================================================

void 
Grid::StoreSqrAnnulus( SPoint centerPoint, double radius, double epsilon )
{
  int i, j;
  static SPoint outerRingVector ( epsilon + radius , epsilon + radius );

  // get the outer bounding box of the annulus
  SPoint outerFromPt = centerPoint - outerRingVector ;
  SPoint outerToPt = centerPoint + outerRingVector ;

  // get the inner bounding box of the annulus
  SPoint innerFromPt = Transformation::GetPoint( centerPoint, 
					 radius - epsilon, 5*M_PI/4 );
  SPoint innerToPt = Transformation::GetPoint( centerPoint, 
					 radius - epsilon, M_PI/4 );

  // these points should be all valid since the bounding box was adjusted 
  // for the +- epsilon queries 

  Index innerFromIndex = GetIndex( innerFromPt );
  Index innerToIndex = GetIndex( innerToPt );
  Index pIndex = GetIndex( centerPoint );
  Index outerFromIndex = GetIndex( outerFromPt );
  Index outerToIndex = GetIndex( outerToPt );

  // Since we are going to go with a loop through all of these buckets and
  // test whether or not they cover the anulus, we don't need to assert
  // anything here, each bucket that falls between these boundaries will
  // be systematically evaluated whether or not it should be taken --
  // there should not be a chance for repeats

  // just handle the most obvious of the special cases
  if ( pIndex == outerFromIndex && pIndex == outerToIndex ) {
    bucketsInAnnulus.Append( pIndex );
    return; // nothing else to do here
  }
  

  // the general strategy here is the following:
  // -------------------------------------------
  // don't worry about outlier cells, get all of them that are between the
  // outer annulus box and the inner anulus box
  // ---------------------------------------------

  // begin the traversal from the lower outside corner
  for( i = outerFromIndex.x; i <= innerFromIndex.x; i++ ) {
    for( j = outerFromIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );
      bucketsInAnnulus.Append( n );
    }
  }
  
  // from the current position
  for( ; i < innerToIndex.x; i++ ) {
    // we need to take care of two regions, so we have two for-loops
    for( j = outerFromIndex.y; j <= innerFromIndex.y; j++ ) {
      Index n ( i,j );
      bucketsInAnnulus.Append( n );
    }
    
    for( j = innerToIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );
      bucketsInAnnulus.Append( n );
    }
  }
  
  // from the current position
  for( ; i <= outerToIndex.x; i++ ) {
    for( int j = outerFromIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );
      bucketsInAnnulus.Append( n );
    }
  }
}

void 
Grid::StoreLinAnnulusSquarely ( SPoint centerPoint, 
				     double radius, double epsilon )
{
  int i, j;
  SPoint closestComparePt, farthestComparePt, adjustedCenterPt;
  static SPoint outerRingVector ( epsilon + radius , epsilon + radius );

  // get the outer bounding box of the annulus
  SPoint outerFromPt = centerPoint - outerRingVector ;
  SPoint outerToPt = centerPoint + outerRingVector ;

  // get the inner bounding box of the annulus
  SPoint innerFromPt = Transformation::GetPoint( centerPoint, 
					 radius - epsilon, 5*M_PI/4 );
  SPoint innerToPt = Transformation::GetPoint( centerPoint, 
					 radius - epsilon, M_PI/4 );

  // these points should be all valid since the bounding box was adjusted 
  // for the +- epsilon queries 

  Index innerFromIndex = GetIndex( innerFromPt );
  Index innerToIndex = GetIndex( innerToPt );
  Index pIndex = GetIndex( centerPoint );
  Index outerFromIndex = GetIndex( outerFromPt );
  Index outerToIndex = GetIndex( outerToPt );

  // Since we are going to go with a loop through all of these buckets and
  // test whether or not they cover the anulus, we don't need to assert
  // anything here, each bucket that falls between these boundaries will
  // be systematically evaluated whether or not it should be taken --
  // there should not be a chance for repeats

  // just handle the most obvious of the special cases
  if ( pIndex == outerFromIndex && pIndex == outerToIndex ) {
    bucketsInAnnulus.Append( pIndex );
    return; // nothing else to do here
  }
  
  // begin the traversal from the lower outside corner
  for( i = outerFromIndex.x; i <= innerFromIndex.x; i++ ) {
    for( j = outerFromIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );

      if ( IsBucketInAnnulus( n, centerPoint, radius, epsilon ) ) {
	bucketsInAnnulus.Append( n );
      }
    }
  }
    
  // from the current position
  for( ; i < innerToIndex.x; i++ ) {
    // we need to take care of two regions, so we have two for-loops
    for( j = outerFromIndex.y; j <= innerFromIndex.y; j++ ) {
      Index n ( i,j );

      if ( IsBucketInAnnulus( n, centerPoint, radius, epsilon ) ) {
	bucketsInAnnulus.Append( n );
      }
    }

    for( j = innerToIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );

      if ( IsBucketInAnnulus( n, centerPoint, radius, epsilon ) ) {
	bucketsInAnnulus.Append( n );
      }
   }
  }

  // from the current position
  for( ; i <= outerToIndex.x; i++ ) {
    for( int j = outerFromIndex.y; j <= outerToIndex.y; j++ ) {
      Index n ( i,j );

      if ( IsBucketInAnnulus( n, centerPoint, radius, epsilon ) ) {
	bucketsInAnnulus.Append( n );
      }
    }
  }
}


// We are indebted to Drago for the algorithm behind this function
void 
Grid::StoreLinAnnulusLinearly ( SPoint centerPt, double radius, double epsilon )
{
  double start;
  SPoint upperBeginPt ( centerPt.x, centerPt.y + (radius+epsilon) );
  SPoint lowerBeginPt ( centerPt.x, centerPt.y - (radius+epsilon) );
  SPoint rightEndPt ( centerPt.x + (radius+epsilon), centerPt.y );
  SPoint leftEndPt ( centerPt.x - (radius+epsilon), centerPt.y );

  // get the indices of these points
  Index upperBeginIndex = GetIndex( upperBeginPt );
  Index lowerBeginIndex = GetIndex( lowerBeginPt );
  Index rightEndIndex = GetIndex( rightEndPt );
  Index leftEndIndex = GetIndex( leftEndPt );


  // first we go through the 1st quadrant just to collect points
  // which we'll flip later to cover the other 3 quadrants of the annulus
  // and only then, we'll collect the buckets
  // ...
  // yeah, this would have been a good idea if the centerPoint was
  // always in the middle of a cell, which is NOT the case
  // so we need to just do it quadrant by quadrant
  // ...
  // yeah !!! but now this is exactly the case, so... maybe we can do this
  // ???


  // Quadrant 1
  //cout << "Quadrant 1" << endl;
  start = upperBeginIndex.y ; 

  for( int xDir = upperBeginIndex.x ; xDir <= rightEndIndex.x; xDir++ ) {
    int yDir = start; 
    assert( yDir >= rightEndIndex.y );
    
    // skip the possibly outlying initial buckets
    while( yDir >= rightEndIndex.y ) {
      Index n ( xDir, yDir );

      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	start = yDir; // next time start from here
	break; // just advance the index until it covers the annulus
      } 

      // handle the annulus going out of the grid
      else if ( n.x == upperBeginIndex.x ) {
	for( ; xDir < rightEndIndex.x &&
	       ! IsBucketInAnnulus( n, centerPt, radius, epsilon );
	     xDir ++, n.x ++ )
	  ; // end of for loop do nothing, just go until you find annulus 
	break;
      }

      yDir -- ;
    }
    
    // collect the buckets in the annulus
    while ( yDir >= rightEndIndex.y ) {
      Index n ( xDir, yDir );
  
      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	bucketsInAnnulus.Append( n );
      }
      else
	break;
      
      yDir --;
    }
  }

  // Quadrant 2
  //cout << "Quadrant 2" << endl;

  start = upperBeginIndex.y ; 

  for( xDir = upperBeginIndex.x - 1 ; xDir >= leftEndIndex.x; xDir -- ) {
    int yDir = start; 
    assert( yDir >= leftEndIndex.y );

    // skip the possibly outlying initial buckets
    while( yDir >= rightEndIndex.y ) {
      Index n ( xDir, yDir );

      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	start = yDir; // next time start from here
	break; // just advance the index until it covers the annulus
      }      

      // handle the annulus going out of the grid
      else if ( n.x == upperBeginIndex.x ) {
	for( ; xDir > leftEndIndex.x &&
	       ! IsBucketInAnnulus( n, centerPt, radius, epsilon );
	     xDir -- , n.x -- )
	  ; // end of for loop. do nothing, just go until annulus is found
	break;
      }

      yDir -- ;
    }

    // collect the buckets in the annulus
    while ( yDir >= leftEndIndex.y ) {
      Index n ( xDir, yDir );
     
      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	bucketsInAnnulus.Append( n );
      }
      else
	break;
      
      yDir --;
    }
  }


  // Quadrant 3
  //cout << "Quadrant 3" << endl;

  start = lowerBeginIndex.y ; 

  for( xDir = lowerBeginIndex.x ; xDir >= leftEndIndex.x; xDir -- ) {
    int yDir = start; 
    assert( yDir <= leftEndIndex.y );
    
    // skip the possibly outlying initial buckets
    while( yDir <= leftEndIndex.y ) {
      Index n ( xDir, yDir );

      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	start = yDir; // next time start from here
	break; // just advance the index until it covers the annulus
      }      

      // handle the annulus going out of the grid
      else if ( n.x == lowerBeginIndex.x ) {
	for( ; xDir > leftEndIndex.x &&
	       ! IsBucketInAnnulus( n, centerPt, radius, epsilon );
	     xDir -- , n.x -- )
	  ; // end of for loop. do nothing, just go until annulus is found
	break;
      }

      yDir ++ ;
    }

    // collect the buckets in the annulus
    while ( yDir < leftEndIndex.y ) {
      Index n ( xDir, yDir );
     
      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	bucketsInAnnulus.Append( n );
      }
      else
	break;
      
      yDir ++ ;
    }
  }


  // Quadrant 4
  //  cout << "Quadrant 4" << endl;

  start = lowerBeginIndex.y ; 

  for(  xDir = lowerBeginIndex.x + 1 ; xDir <= rightEndIndex.x; xDir ++ ) {
    int yDir = start; 
    assert( yDir <= leftEndIndex.y );

   // skip the possibly outlying initial buckets
    while( yDir <= leftEndIndex.y ) {
      Index n ( xDir, yDir );

      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	start = yDir; // next time start from here
	break; // just advance the index until it covers the annulus
      }      

      // handle the annulus going out of the grid
      else if ( n.x == lowerBeginIndex.x ) {
	for( ; xDir < rightEndIndex.x &&
	       ! IsBucketInAnnulus( n, centerPt, radius, epsilon );
	     xDir ++, n.x ++ )
	  ; // end of for loop do nothing, just go until you find annulus 
	break;
      }

      yDir ++ ;
    }

    // collect the buckets in the annulus
    while ( yDir < leftEndIndex.y ) {
      Index n ( xDir, yDir );
     
      if ( IsBucketInAnnulus( n, centerPt, radius, epsilon ) ) {
	bucketsInAnnulus.Append( n );
      }
      else
	break;
      
      yDir ++ ;
    }
  }

}

void
Grid::PreStampAnnulus( double radius, double epsilon ) 
{
  // the stamp is w.r.t the center of index (0,0)
  SPoint centerPoint ( ptBox.min.x + cellSizeOnX / 2.0 ,
		      ptBox.min.y + cellSizeOnY / 2.0 ); 

  // enlarge the effective epsilon to offset the approximation of the
  // centerPoint which could be anywhere in the cell) with 
  // the center of a cell
  assert( cellSizeOnX == cellSizeOnY );
  double correctedEpsilon = epsilon + ( cellSizeOnX * M_SQRT2 / 2.0 );

#if ( GET_BUCKETS_FASTSQR && ! GET_BUCKETS_LINEARLY )
  StoreSqrAnnulus( centerPoint, radius, correctedEpsilon );
#elif ( ! GET_BUCKETS_FASTSQR && ! GET_BUCKETS_LINEARLY )
  StoreLinAnnulusSquarely( centerPoint, radius, correctedEpsilon );
#elif ( ! GET_BUCKETS_FASTSQR && GET_BUCKETS_LINEARLY )
  StoreLinAnnulusLinearly( centerPoint, radius, correctedEpsilon );
#else
  assert( !"PreStampAnnulus needs a method of bucket retrieval");
#endif
}

void inline
Grid::GetPreStampedBuckets( const SPoint &centerPoint, 
			     List<List<SPoint>*> *allBuckets )
{
  Index centerPtBucket = GetIndex( centerPoint );

  ListIterator<Index> bucketIterator( &bucketsInAnnulus ); 
  for (; !bucketIterator.IsDone(); bucketIterator.Next()) {
    Index n = bucketIterator.Item() + centerPtBucket;
    if ( IndexIsOK( n ) )
      allBuckets->Append( grid[ n.x ][ n.y ] );
  }
}


// ==================================================================
#endif        // MESH, STAMP_ANNULUS
// ==================================================================



// private helper to GetBucketsInAnulus
inline double
Grid::ClosestWall( double norm, double smaller, double bigger ) const
{
  assert( smaller < bigger );

  if ( smaller < norm && bigger <= norm ) 
    return bigger;
  
  else if ( smaller >= norm && bigger > norm ) 
    return smaller;
  
  else { // the "in between" case 
    return ( ( bigger - norm <= norm - smaller )? bigger: smaller );
  }
}

// private helper to GetBucketsInAnulus
inline double
Grid::FarthestWall( double norm, double smaller, double bigger ) const
{
  assert( smaller < bigger );

  if ( smaller < norm && bigger <= norm ) 
    return smaller;
  
  else if ( smaller >= norm && bigger > norm ) 
    return bigger;
  
  else { // the "in between" case 
    return ( ( bigger - norm <= norm - smaller )? smaller: bigger );
  }
}

// private helper to GetBucketsInAnulus
inline double
Grid::ClosestBetweenWall( double norm, double smaller, double bigger ) const
{
  assert( smaller < bigger );

  if ( smaller < norm && bigger <= norm ) 
    return bigger;
  
  else if ( smaller >= norm && bigger > norm ) 
    return smaller;
  
  else // the "in between" case 
    return norm;
}

// private helper to GetBucketsInAnulus
inline double
Grid::FarthestBetweenWall( double norm, double smaller, double bigger ) const
{
  assert( smaller < bigger );

  if ( smaller < norm && bigger <= norm ) 
    return smaller;
  
  else if ( smaller >= norm && bigger > norm ) 
    return bigger;
  
  else  // the "in between" case 
    return norm;
}

// tests if the bucket at index n lies in the annulus determined by 
// centerPoint and radius +- epsilon. if so, return true else false.
//
// the general strategy here is the following:
// -------------------------------------------
// * if the bucket's closest corner (to the central point) is IN the
// annulus, take the bucket 
// * if not, if the bucket's farthest corner (w.r.t. the central point)
// is IN the annulus, take the bucket
// * if not, then both corners are out of the annulus and the only way
// the square is still covering it is that the closer corner needs to be
// BELOW the annulus AND the farther point needs to be ABOVE the annulus
// --------------------------------------------
// In the "in-between" cases, do the same thing but not on the corners
// but on the points obtained in the following way:
//    for x-direction's closestPt: 
//                < adjustedCenterPt.x, current cell's closestWall >
//    for x-direction's farthestPt:
//                < adjustedCenterPt.x, current cell's farthestWall >
// The same should be done for the y-direction.
// ---------------------------------------------
inline bool
Grid::IsBucketInAnnulus( Index n, SPoint centerPoint, 
			 double radius, double epsilon ) const
{
  double left, right, up, down;
  SPoint closestComparePt, farthestComparePt, adjustedCenterPt;
  double outerDistSquared = SQR( radius + epsilon );
  double innerDistSquared = SQR( radius - epsilon );
  double shorterDistSqrd, longerDistSqrd;

#if ( ! STAMP_ANNULUS )
  // these points should be all valid since the bounding box was adjusted 
  // for the +- epsilon queries
  assert( PointIsOK( centerPoint ) );
  assert( IndexIsOK( n ) ); 
#endif
  
  // these are the current cell's walls w.r.t. 0,0 
  left = n.x * cellSizeOnX;
  right = (n.x + 1) * cellSizeOnX;
  down = n.y * cellSizeOnY;
  up = (n.y + 1) * cellSizeOnY;
  
  // need to adjust the point to be comparable to the cell coordinates
  adjustedCenterPt = centerPoint - ptBox.min;


  // when we test for inclusion within the outer annulus limits,
  // we should look at the walls which are closer to the point
  // the "in-between" case is handled automatically by ClosestWall too.
  closestComparePt.x = ClosestWall( adjustedCenterPt.x, left, right );
  closestComparePt.y = ClosestWall( adjustedCenterPt.y, down, up );
  shorterDistSqrd = Transformation::LengthSQR( adjustedCenterPt,
					       closestComparePt );

  // when we test for inclusion within the inner annulus limits
  // we should look at the walls which are farthest from the central point
  // the "in-between" case is handled automatically by FarthestWall too.
  farthestComparePt.x = FarthestWall( adjustedCenterPt.x, left, right );
  farthestComparePt.y = FarthestWall( adjustedCenterPt.y, down, up );
  longerDistSqrd = Transformation::LengthSQR( adjustedCenterPt,
					      farthestComparePt );

  // if the point is "in-between", there are a few things possible
  if ( ( left < adjustedCenterPt.x && adjustedCenterPt.x < right ) ||
       ( down < adjustedCenterPt.y && adjustedCenterPt.y < up ) ) {

    // if the center point is in the index n, check if the farthest point 
    // is more than the inner radius of the annulus
    if ( left < adjustedCenterPt.x && adjustedCenterPt.x < right &&
	 down < adjustedCenterPt.y && adjustedCenterPt.y < up &&
	 longerDistSqrd >= innerDistSquared ) {
      return true;
    }

    // check if the center point is "in-between" the x- or y-walls of index n
    // but the anulus does not cross the diagonal between the
    // closest and furthest point ( the circle intersects just the 
    // closest wall of the cell without including any of its corners )
    if ( outerDistSquared <= shorterDistSqrd ) {
      closestComparePt.x = ClosestBetweenWall( adjustedCenterPt.x, 
					       left, right );
      closestComparePt.y = ClosestBetweenWall( adjustedCenterPt.y, 
					       down, up );
      shorterDistSqrd = Transformation::LengthSQR( adjustedCenterPt,
						   closestComparePt );
      farthestComparePt.x = FarthestBetweenWall( adjustedCenterPt.x, 
						 left, right );
      farthestComparePt.y = FarthestBetweenWall( adjustedCenterPt.y, 
						 down, up );
      longerDistSqrd = Transformation::LengthSQR( adjustedCenterPt,
						  farthestComparePt );
    }
  }  
    

  // one of the two points must be in the annulus or it could be that
  // the closest is less than the "between" distance and the farthest
  // is more than that distance (the annulus must be in the middle)
  if ( Transformation::IsLengthBetweenSQR( shorterDistSqrd,
					   innerDistSquared, 
					   outerDistSquared ) ||
       Transformation::IsLengthBetweenSQR( longerDistSqrd,
					   innerDistSquared, 
					   outerDistSquared ) || 
       ( shorterDistSqrd < innerDistSquared && 
	 longerDistSqrd > outerDistSquared ) ) { 
    return true;
  }
     
  
  // the index n did not lie on the annulus
  return false;
}






// =========================  ALIGNMENT GRID  ========================


// Making an alignment grid 
AlignmentGrid::AlignmentGrid( BoundingBox _box, double _epsilon, 
			      List<SPoint> *imagePoints )
{
  box = _box;
  epsilon = _epsilon;

  assert( AlignGridFactor > 0 );
  cellSizeOnX = cellSizeOnY = epsilon/AlignGridFactor;

  // maybe we don't need these values as class vars if we can 
  // always compute them from the box 
  areaSizeOnX = box.max.x - box.min.x;
  areaSizeOnY = box.max.y - box.min.y;

#if DEBUG
  cout << "Making an alignment grid... " << endl;
  cout << "\tThe alignGridFactor: " << AlignGridFactor << 
    ", epsilon: " << epsilon << endl; 
#endif
    
  grid = new Grid( 0, box, cellSizeOnX, cellSizeOnY);

#if DEBUG
  cout << "\tEntering all points in the alignment grid ... ";
#endif

  grid->InitializeWith( imagePoints );

#if DEBUG
  cout << "Done." << endl << endl;
#endif

}

AlignmentGrid::~AlignmentGrid()
{
  delete grid;
}


#if ( EFFECTIVE_EPSILON && ! BIT_GRID )
bool
AlignmentGrid::Align( List<SPoint> *modelPoints, Transformation t ) const
{
  //tempEpsilons will hold the min epsilons until a match is found
  assert( tempEpsilons == NULL );
  tempEpsilons = new List<double>;
  
  List< List<SPoint>* > buckets;
  ListIterator<SPoint> pointIterator( modelPoints );
  for (; !pointIterator.IsDone(); pointIterator.Next()) {
    SPoint thePoint = pointIterator.Item();

    thePoint = t.Transform( thePoint );
    //cout << "Trying to align " << thePoint << endl;

    grid->GetFBuckets(thePoint, &buckets );

    if ( ! HasAtLeastOneMatch( thePoint, &buckets ) ) {
      delete tempEpsilons; 
      tempEpsilons = NULL;
      return false;
    }
  }

  minEpsilons.Append( tempEpsilons );
  tempEpsilons = NULL;

  return true;
}

bool
AlignmentGrid::HasAtLeastOneMatch( SPoint thePoint,
                                   List< List<SPoint>* > *buckets ) const
{
  static double eps2 = epsilon*epsilon;
  assert( tempEpsilons );
  double minDistSQR = -1;

  ListIterator< List<SPoint>* > bucketIterator( buckets );
  for (; !bucketIterator.IsDone(); bucketIterator.Next()) {
    ListIterator<SPoint> possibleMatches(bucketIterator.Item());
    for (; ! possibleMatches.IsDone(); possibleMatches.Next()) {
      SPoint matchPt = possibleMatches.Item();
      double distance = Transformation::LengthSQR( thePoint, matchPt );

      // make sure this is a real match
      if ( Transformation::IsLengthBetweenSQR( distance, 0, eps2 ) ) {
	if ( minDistSQR > distance || minDistSQR < 0 )
	  minDistSQR = distance;
      }

    }
  }

  if (minDistSQR >= 0 )
    tempEpsilons->Append( sqrt( minDistSQR ) );

  return ( minDistSQR >= 0 );
}


#elif ( ! EFFECTIVE_EPSILON && ! BIT_GRID )

bool
AlignmentGrid::Align( List<SPoint> *modelPoints, Transformation t ) const
{
  List< List<SPoint>* > buckets;
  ListIterator<SPoint> pointIterator( modelPoints );
  for (; !pointIterator.IsDone(); pointIterator.Next()) {
    SPoint thePoint = pointIterator.Item();

    thePoint = t.Transform( thePoint );
    //cout << "Trying to align " << thePoint << endl;

    grid->GetFBuckets(thePoint, &buckets);

    if ( ! HasAtLeastOneMatch( thePoint, &buckets ) ) {
      //cout << thePoint << " had no match." << endl;
      return false;
    }
  }
  return true;
}

bool 
AlignmentGrid::HasAtLeastOneMatch( SPoint thePoint, 
				   List< List<SPoint>* > *buckets ) const
{
  static double eps2 = epsilon*epsilon;

#if STATS
  int count= 0;
  static int callnum = 0;
  callnum++;
#endif
  
  ListIterator< List<SPoint>* > bucketIterator( buckets );
  for (; !bucketIterator.IsDone(); bucketIterator.Next()) {
    ListIterator<SPoint> possibleMatches(bucketIterator.Item());
    for (; ! possibleMatches.IsDone(); possibleMatches.Next()) {
      SPoint matchPt = possibleMatches.Item();

#if DEBUG
      //cout << "The possible match for " << thePoint << " is " << matchPt
      //   << endl;
#endif
#if STATS
      count++;
#endif
      
      // make sure this is a real match
      if ( Transformation::IsLengthBetweenSQR( thePoint, matchPt, 0, eps2 ) ) {
#if STATS
	//          cout << "call " << callnum << " did " << count <<
	//    " iterations before it got a match." << endl;
#endif	
	return true;
      }
      
    }
  }

#if STATS
  // cout << "call " << callnum << " did " << count << 
  // " iterations before it failed." << endl;
#endif

  return false;
}
#endif // EFFECTIVE_EPSILON && ! BIT_GRID test



// =========================  ALIGNMENT BIT GRID  ========================


// Making an alignment grid 
AlignmentBitGrid::AlignmentBitGrid( BoundingBox _box, double _epsilon, 
				    List<SPoint> *imagePoints )
{
  box = _box;
  epsilon = _epsilon;

  assert( AlignGridFactor > 0 );
  cellSizeOnX = cellSizeOnY = epsilon / AlignGridFactor;

  areaSizeOnX = box.max.x - box.min.x;
  areaSizeOnY = box.max.y - box.min.y;

  // we take the floor of +1 and not ceil because if the division 
  // results in an integer, then we are in trouble -- a point lying
  // on the boundary will get assigned an index outside of the grid
  elementsOnX = floor( ( areaSizeOnX / cellSizeOnX ) + 1 );
  elementsOnY = floor( ( areaSizeOnY / cellSizeOnY ) + 1 );

#if DEBUG
  cout << "Making an alignment bit grid... " << endl;
  cout << "\tThe alignGridFactor: " << AlignGridFactor << 
    ", epsilon: " << epsilon << endl; 
  cout << "\tThe bounding box is " << box << endl;
  cout << "\tThe cell size on X and Y are respectively " << cellSizeOnX <<
    ", " << cellSizeOnY << endl;
  cout << "\tThe area size on X and Y are respectively " << areaSizeOnX <<
    ", " << areaSizeOnY << endl;
  cout << endl << "\tAllocating a bit grid with size " << elementsOnX
    << " by " << elementsOnY << " ...";
#endif

  // allocate the dynamic 2-D array of chars
  grid = new char* [ elementsOnX ];
  for (int i=0; i < elementsOnX; i++) {
    grid[i] = new char [ elementsOnY ];
    memset( grid[i], 0, elementsOnY * sizeof( char ) );
  }

#if DEBUG
  cout << "Done." << endl;
  cout << "\tEntering all points in the alignment bit grid ... ";
#endif

  InitializeWith( imagePoints );

#if DEBUG
  cout << "Done." << endl << endl;
#endif
}

AlignmentBitGrid::~AlignmentBitGrid()
{
  for (int i=0; i < elementsOnX; i++) {
    delete grid[i];
  }
  delete grid;
}

void 
AlignmentBitGrid::InitializeWith(List<SPoint> *imagePoints)
{
  ListIterator<SPoint> pointIterator( imagePoints ); 
  for (; !pointIterator.IsDone(); pointIterator.Next()) {
    SPoint thePoint = pointIterator.Item();
    SpreadPointInGrid( thePoint );

#if DEBUG 
    //      cout << "Inserting/Spreading the SPoint " << thePoint << 
    //" in the alignment bit grid." << endl;
#endif      
  }
}


inline
double
AlignmentBitGrid::ClosestWall( double norm, 
			       double smaller, double bigger ) const
{
  assert( smaller < bigger );

  if ( smaller < norm && bigger <= norm ) 
    return bigger;
  
  else if ( smaller >= norm && bigger > norm ) 
    return smaller;
  
  else // the "in between" case 
    assert( !"We should not be here if the in-between cases were handled");
    return norm;
}

void
AlignmentBitGrid::SpreadPointInGrid( SPoint p )
{
  double left, right, up, down;
  SPoint comparePt, adjustedPt;
  static double epsilonSquared = SQR( epsilon );
  SPoint epsilonVector ( epsilon, epsilon );

  assert( PointIsOK( p ) );

  // these points should be all valid since the bounding box was adjusted 
  // for the +- epsilon queries 
  SPoint fromPt = p - epsilonVector;
  SPoint toPt = p + epsilonVector; 
// MODIFIFICATION (SURESH) Sat Nov 21 20:34:58
  if( !( PointIsOK( fromPt ) && PointIsOK( toPt ) ) ){
      cout << fromPt << endl;
      cout << toPt << endl;
      cout << box.min << endl;
      cout << box.max << endl;
      return;
  }
  
  
  Index pIndex = GetIndex( p );
  Index fromPtIndex = GetIndex( fromPt );
  Index toPtIndex = GetIndex( toPt );
  assert( ( fromPtIndex != toPtIndex ) || 
	  ( pIndex == fromPtIndex && pIndex == toPtIndex ) );
  assert( fromPtIndex.x <= pIndex.x && pIndex.x <= toPtIndex.x && 
	  fromPtIndex.y <= pIndex.y && pIndex.y <= toPtIndex.y );

  // mark the cell where the point is
  grid [ pIndex.x ] [ pIndex.y ] = Marked;

  if ( pIndex == fromPtIndex && pIndex == toPtIndex )
    return; // nothing else to do here

  // handle the "in between" cases
  // it saves just a tiny bit -- a few mults and comps and SQR
  for( int x = fromPtIndex.x; x <= toPtIndex.x; x++ ) {
    grid [ x ] [ pIndex.y ] = Marked;      // at 0 and pi
  }
  for( int y = fromPtIndex.y; y <= toPtIndex.y; y++ ) {
    grid [ pIndex.x ] [ y ] = Marked;      // at pi/2 and 3*pi/2
  }
  

  for( int i = fromPtIndex.x; i <= toPtIndex.x; i++ ) {
    for( int j = fromPtIndex.y; j <= toPtIndex.y; j++ ) {
      Index n ( i,j );

      // these points should be all valid since the bounding box was adjusted 
      // for the +- epsilon queries 
      assert( IndexIsOK( n ) ); 

      // don't bother with any calculations if this index was already marked
      if ( grid[ n.x ][ n.y ] == Marked ) 
	continue;
      
      // these are the current cell's walls
      left = n.x * cellSizeOnX;
      right = (n.x + 1) * cellSizeOnX;
      down = n.y * cellSizeOnY;
      up = (n.y + 1) * cellSizeOnY;

      // need to adjust the point to be comparable to the cell coordinates
      adjustedPt = p - box.min;

    
      // when we test for inclusion within the outer annulus limits,
      // we should look at the walls which are closer to the point
      // the "in-between" case is handled automatically by ClosestWall --
      // it returns the adjustedCenterPt coordinate if this is the case 
      comparePt.x = ClosestWall( adjustedPt.x, left, right );
      comparePt.y = ClosestWall( adjustedPt.y, down, up );

      // if we are here, we should have the closest point of the
      // current cell 'n' in the point 'comparePt'

      if ( Transformation::IsLengthBetweenSQR( comparePt, adjustedPt,
					       0, epsilonSquared ) ) {
	grid[ n.x ][ n.y ] = Marked;
      }
    }
  }
}  



bool 
AlignmentBitGrid::Align( List<SPoint> *modelPoints, Transformation t ) const
{
  ListIterator<SPoint> pointIterator( modelPoints ); 
  for (; !pointIterator.IsDone(); pointIterator.Next()) {
    SPoint thePoint = pointIterator.Item();

    thePoint = t.Transform( thePoint );
    //cout << "Trying to align in bit grid " << thePoint << endl;
    
    if ( ! PointIsOK( thePoint ) )
      return false;

    Index n = GetIndex( thePoint );
    assert( IndexIsOK( n ) );

    if ( grid [ n.x ] [ n.y ] == Not_Marked ) {
      //cout << thePoint << " had no match." << endl;
      return false;
    }
  }
  return true;
}

inline Index 
AlignmentBitGrid::GetIndex(SPoint p) const
{
  assert( PointIsOK(p) );
  SPoint adjustedPt = p - box.min;
  Index n(floor(adjustedPt.x / cellSizeOnX), 
	  floor(adjustedPt.y / cellSizeOnY));
#if DEBUG
  //cout << "SPoint " << p << " fits in bucket " << n << endl;
#endif
  return n;
}

bool
AlignmentBitGrid::PointIsOK(SPoint p) const
{
  return (p.x >= box.min.x && p.x <= box.max.x && 
	  p.y >= box.min.y && p.y <= box.max.y);
}

bool
AlignmentBitGrid::IndexIsOK(Index n) const
{
  return (n.x >= 0 && n.x < elementsOnX &&
	  n.y >= 0 && n.y < elementsOnY);
}

// =========================  SIMPLE MESH  ========================

#if ( SIMPLE_MESH )

SimpleMesh::SimpleMesh(BoundingBox _box, double _radius, 
		       double _epsilon, double _beta, 
		       List<SPoint> *imagePoints) 
{
  epsilon = _epsilon;
  beta = _beta;
  radius = _radius;
  box = _box;

  assert( MeshFactor > 0 );
  cellSizeOnX = cellSizeOnY = epsilon * MeshFactor;

  areaSizeOnX = box.max.x - box.min.x;
  areaSizeOnY = box.max.y - box.min.y;

#if DEBUG
  cout << "Creating a simple mesh...." << endl;
  cout << "\tThe bounding box is " << box << endl;
  cout << "\tThe radius is " << radius << endl;
  cout << "\tEpsilon: " << epsilon << ", beta: " << beta <<
    ", MeshFactor: " << MeshFactor << endl;
#endif
  
  grid = new Grid( 0, box, cellSizeOnX, cellSizeOnY);
  

#if DEBUG
  cout << "\tEntering all points in the simple mesh ... ";
#endif

  grid->InitializeWith( imagePoints );

#if ( STAMP_ANNULUS )
  grid->PreStampAnnulus( radius, epsilon );
#endif

#if DEBUG
  cout << "Done." << endl << endl;
#endif
}



SimpleMesh::~SimpleMesh() 
{
  delete grid;
}


// it returns dynamically alolocated list of lists -- this way, 
// we don't need to make a copy of the list and incur a big cost 
void
SimpleMesh::FindPossiblePairEnds(SPoint centerPoint, 
			   List< List<SPoint>* > *allBuckets) const
{
#if ( STAMP_ANNULUS )
  grid->GetPreStampedBuckets( centerPoint, allBuckets );

#else // !STAMP_ANNULUS

#  if ( GET_BUCKETS_FASTSQR && ! GET_BUCKETS_LINEARLY )
  grid->GetBucketsInAnulusFast( centerPoint, radius, epsilon, allBuckets );
#  elif ( ! GET_BUCKETS_FASTSQR && ! GET_BUCKETS_LINEARLY )
  grid->GetBucketsInAnulus( centerPoint, radius, epsilon, allBuckets );
#  elif ( ! GET_BUCKETS_FASTSQR && GET_BUCKETS_LINEARLY )
  grid->GetBucketsInAnulusLinearly( centerPoint, radius, epsilon, allBuckets );
#  else
  assert( !"FindPossiblePairEnds needs a method of bucket retrieval");
#  endif

#endif // STAMP_ANNULUS

#if DEBUG
  //cout << "Got " << allBuckets->NumInList() << " buckets." << endl;
#endif
}

#endif // SIMPLE_MESH

// =============================  MESH  ========================


Mesh::Mesh(BoundingBox _box, double _radius, 
	   double _epsilon, double _beta, List<SPoint> *imagePoints) 
{
  epsilon = _epsilon;
  beta = _beta;
  radius = _radius;
  box = _box;

  cellSizeOnX = 2.0* epsilon + beta;
  cellSizeOnY = 2.0*( sqrt( SQR(radius + epsilon) - 
			    SQR(radius - epsilon - beta)));

  // maybe we don't need these values as class vars if we can 
  // always compute them from the box 
  areaSizeOnX = box.max.x - box.min.x;
  areaSizeOnY = box.max.y - box.min.y;

  // this is the theta that guarantees the proper coverage of the annulus
  // with the cells of size as defined above
  rotationalStep = 2.0 * acos((radius - beta - epsilon)/(radius -
							 epsilon));

  // take the CEILING to be safe but we might check twice some pairs...
  double nSteps = 2.0 * M_PI / rotationalStep;
  numberOfSteps = ceil( nSteps ); 

  // precompute the sample points to avoid calculating them |image| times
  samplePoints = new SPoint [ numberOfSteps ];
  for (int j = 0; j < numberOfSteps; j++) {
    double rotationalAngle = j * rotationalStep;
    SPoint center (0,0);
    samplePoints[j] = Transformation::GetPoint( center, 
						radius, rotationalAngle);
  }  

#if DEBUG
  cout << "Creating a mesh...." << endl;
  cout << "\tThe bounding box is " << box << endl;
  cout << "\tThe radius is " << radius << endl;
  cout << "\tThe epsilon and beta are " << epsilon << " and " << beta <<
    endl;
  cout << "\tThe cell size on X and Y are respectively " << cellSizeOnX <<
    ", " << cellSizeOnY << endl;
  cout << "\tThe area size on X and Y are respectively " << areaSizeOnX <<
    ", " << areaSizeOnY << endl;
  cout << "\tThe rotational step is " << rotationalStep << endl;
  cout << "\tThe number of steps needed to cover the annulus are " <<
    numberOfSteps << endl;

  cout << endl << "\tAllocating the Grids now (" << numberOfSteps
       << " of them)" << endl;
#endif
  
  // allocate the dynamic array of Grids
  grids = new Grid* [numberOfSteps];
  for (int i=0; i < numberOfSteps; i++) {
    grids[i] = new Grid( i * rotationalStep, box, cellSizeOnX, cellSizeOnY);
  }

#if DEBUG
  cout << "\tEntering all points in the mesh ... ";
#endif
#if STATS 
  cout << grids[0]->GetElements() << " ";
#endif
  InitializeWith( imagePoints );

#if DEBUG
  cout << "Done." << endl << endl;
#endif
}



Mesh::~Mesh() 
{
  for (int i=0; i < numberOfSteps; i++) {
    delete grids[i];
  }
  delete grids;

  delete samplePoints;
}

// --------------------
// Mesh::Initialize
// --------------------
// to reproduce the effect of the grid rotation with an angle +theta
// (counterclockwise) w.r.t. (0,0), we rotate the points with an andle
// -theta (clockwise) w.r.t. (0,0) and we insert them in the appropriate
// grid (corresponding on the i-th step)....
void
Mesh::InitializeWith(List<SPoint> *allImagePoints)
{
  for (int i = 0; i < numberOfSteps; i++) {
    // the grid is smart, it will rotate the points 
    grids[i]->InitializeWith( allImagePoints );
  }
}

// it returns dynamically alolocated list of lists -- this way, 
// we don't need to make a copy of the list and incur a big cost 
void
Mesh::FindPossiblePairEnds(SPoint &centerPoint, 
			   List< List<SPoint>* > *allBuckets) const
{
  for (int i = 0; i < numberOfSteps; i++) {
    SPoint samplePt = centerPoint + samplePoints[ i ];

    // the grid is smart, it will rotate the point 
    grids[i]->Get4Buckets( samplePt, allBuckets );
  }

#if DEBUG
  //cout << "Got " << allBuckets->NumInList() << " buckets." << endl;
#endif

}



void
Mesh::SelfTest()
		
{
  srand(time(NULL));
  cout << "Running the Mesh SelfTest()..." << endl;
  SPoint min(0,0), max(650, 650);
  BoundingBox box(min, max);
  cout << "Making a mesh..." << endl;
  Mesh theMesh(box, 350.0, 1, 1, (new List<SPoint>));
  cout << "made a local mesh ..." << endl;

  cout << "now testing GetPoint and RotatePoint ..." << endl;

  // this is only testing the GetPoint and RotatePoint behavior
  SPoint center(0,0), p1(0,0), p2(0,0), q(0,0);
  double radius = GetRandomValue(3, 100);
  double theta = GetRandomValue(-2*M_PI, 2*M_PI);

  p1 = Transformation::GetPoint( center, radius, theta );
  
  cout << "GetPoint returned with radius = " << radius << " and theta = " 
       << theta << " the point " << p1 << endl;

  q.x = radius; // where p1 would have been before GetPoint rotated it
  p2 = Transformation::Rotate( p1, -theta );
  SPoint err = p2 - q;

  cout << "Rotate the opposite direction returned point " << p2 << endl;
  cout << "The error in Rotate(GetPoint()) is " << err << endl;

  cout << "End of Mesh test." << endl;
}
  
// -------------------------------------------------
// Algorithm_One::Alogorithm_One
// -------------------------------------------------
// the constructor of our algorithm

Algorithm_One::Algorithm_One(double _epsilon, double _beta) {
  epsilon = (_epsilon > 0.0)? _epsilon: DefaultEpsilon;
  beta = (_beta > 0.0)? _beta: DefaultBeta;
  
  // initialize the rest which needs to be set up later ...
  areaSizeX = areaSizeY = 0.0;
}

Algorithm_One::~Algorithm_One() {
  // do nothing in the destructor for now since all of the fields 
  // are automatic variables (no explicit memory allocation)
}

// -------------------------------------------------
// Algorithm_One::Match
// -------------------------------------------------
// the meat of the algorithm -- it takes 2 sets of points (a model and an
// image) and returns the mapping of the model set into the image set
// the class FormattedResult is TBD later.
  
List<FormattedResult> *
Algorithm_One::Match( List<SPoint> *_pointsInModel, 
		      List<SPoint> *_pointsInImage )
{
    double radius;
    
#if STATS
  int potentialMatches = 0, falsePotentialMatches = 0;
#endif
  List<FormattedResult> *resultList = new List<FormattedResult>;
  List<Transformation> seenBefore;
  Transformation modelToOldCoord, imageToOldCoord;

  modelToOldCoord = PreprocessPoints( _pointsInModel, &pointsInModel, 
				      &boxModel, &largeModelPairs );
#if DEBUG
  cout << "To fit in the new bounding box the model was transformed by "
       << modelToOldCoord << endl;
#endif

  imageToOldCoord = PreprocessPoints( _pointsInImage, &pointsInImage, 
				      &boxImage );
#if DEBUG
  cout << "To fit in the new bounding box the image was transformed by "
       << imageToOldCoord << endl;
#endif

#if ( BIT_GRID )
  AlignmentBitGrid alignmentGrid ( boxImage, epsilon, &pointsInImage ); 
#else
  AlignmentGrid alignmentGrid ( boxImage, epsilon, &pointsInImage ); 
#endif
 
  ListIterator<Pair> largePairIterator( &(largeModelPairs) );
  for (; !largePairIterator.IsDone(); largePairIterator.Next()) {
    Pair pairInModel = largePairIterator.Item();

#if DEBUG
    cout << "Found " << largeModelPairs.NumInList() << 
      " large pairs in the model." << endl;
    cout << "Working with model pair " << pairInModel;
#endif

    // the diameter will be used as a radius in the
    // attempt to discover possible matches
    radius = GetDistance( pairInModel );
    // needed for speed up in IsLengthBetweenSQR which takes square lengths
    double rpe = SQR(radius+epsilon);
    double rme = SQR(radius-epsilon);

#if DEBUG
    cout << " with radius " << radius << " ." << endl << endl;
#endif
    if( radius-epsilon-beta <= 0)
    {
        cout << "TRIVIAL MATCH: Align to ANY point in Q\n";
        exit(0);
    }
    
    // this would take O(sqrt(radius) * n)
#if ( SIMPLE_MESH && ! TRIVIAL )
    SimpleMesh theMesh(boxImage, radius, epsilon, beta, &pointsInImage);
#elif ( ! SIMPLE_MESH && ! TRIVIAL )
    Mesh theMesh(boxImage, radius, epsilon, beta, &pointsInImage);
#elif ( TRIVIAL )
    // do nothing for the trivial "get-every-pair" solution
    cout << "1 ";
#else
    assert( !"Need to compile with the  proper option" );
#endif
    
    ListIterator<SPoint> allPointsInImage( &pointsInImage );

    for (; !allPointsInImage.IsDone(); allPointsInImage.Next()) {
      SPoint imagePoint = allPointsInImage.Item();

#if ( ! TRIVIAL )      
      List< List<SPoint>* > allBuckets; 
      theMesh.FindPossiblePairEnds( imagePoint, &allBuckets );
      //cout << "FindPossiblePairEnds found " << allBuckets.NumInList()
      //   << " non-empty eligible buckets." << endl;
      
      ListIterator< List<SPoint>* > bucketIterator( &allBuckets );
      for (; !bucketIterator.IsDone(); bucketIterator.Next()) {
	ListIterator<SPoint> allPairEnds(bucketIterator.Item());
#else // if trivial
        {
	ListIterator<SPoint> allPairEnds( &pointsInImage );
#endif

	for (; !allPairEnds.IsDone(); allPairEnds.Next()) {

	  Pair pairInImage(imagePoint, allPairEnds.Item());
	  
#if STATS	  
	  potentialMatches++;
#endif
	  // get rid of the false matches
	  if ( ! Transformation::IsLengthBetweenSQR( pairInImage, rme, rpe) ) {
#if STATS
	    falsePotentialMatches++;
#endif
	    continue;
	  }

	  Transformation transf (pairInModel, pairInImage);
	  
          // It makes no sense to try to avoid ALL repeated transforms
          // since they are a lot and also very few of them would repeat.
          // The expense of tracking them will more than offset the savings
          // so here I get rid only of succesful transforms repeats

/* SURESH *///if ( seenBefore.IsInList( transf ) ) 
          //continue;
	  
	  if ( alignmentGrid.Align( &pointsInModel, transf ) ) {
#if DEBUG
	    cout << "Found a result with pairInModel: " << pairInModel <<
	      " and pairInImage: " << pairInImage << endl;
#endif
	    // keep track only of the small number of successful transforms
	    seenBefore.Append( transf );
#if WANT_RESULTS	    
	    Transformation originalT = imageToOldCoord.Compose( transf );
	    List<SPoint> *modelResult = originalT.Transform( &pointsInModel ); 
	    FormattedResult result ( transf, modelResult );
#if DEBUG
	    cout << "The result is " << transf << endl << endl;
	    //cout << transf.translateV.x << " " <<  
	    // transf.translateV.y << " " << transf.theta << endl;
#endif

	    resultList->Append( result );
#endif /* WANT_RESULTS */
	  }
	}
      }
    }
  }

//  cout << endl;
#if STATS
  cout << radius << " " << potentialMatches << " " << 
    falsePotentialMatches << " " << 
    ( potentialMatches - falsePotentialMatches) << " " << 
    pointsInImage.NumInList() << " ";
#if ( EFFECTIVE_EPSILON && ! BIT_GRID )
  cout << GetMaxMinEpsilon() << " ";
#endif //EFFECTIVE_EPSILON && ! BIT_GRID
  cout << seenBefore.NumInList() << endl;

//   cout << potentialMatches << " pairs were considered for potential matches " 
//        << "(points retrieved from the mesh)." << endl;
//   cout << falsePotentialMatches << " of them were false alarms " <<
//     "(not within the epsilon range)." << endl;
//   cout << "This amounts to " <<
//       (((float)potentialMatches)/pointsInImage.NumInList()) 
//        << " potential matches and " 
//        << (((float)falsePotentialMatches)/pointsInImage.NumInList())
//        << " false potentials per image point." << endl;
#endif  // STATS

  // at this point, we know that we could not match anything ...
  return resultList;
}




// Do some work on the point sets that will put them in a format with
// which we can work. Notably,
//    * find their bounding rectangle, 
//    * pick all of the points that are on its 4 sides (needed later), 
//    * enlarge the bounding box with epsilon to handle outlier points
//    * enlarge the bounding rectangle into a square, 
//    * enlarge the square by a factor of sqrt(2), 
//    * find the center of that square, set it to coordinates x=0, y=0
//    * change the coordinates of all points
//    * change the coordinates of the points separated into 4 groups
//    * create a list of pairs made up from the points in these 4 groups
//    * return the new points, the bounding box, and the pairs
//    * the return value contains the vector that was used to change the
//        coordinate system of the points
// Yes, it's quite a lot, isn't it ?
Transformation
Algorithm_One::PreprocessPoints(List<SPoint> *originalPts,
				List<SPoint> *newPts,
				BoundingBox *box,
				List<Pair> *largePairs) const
{
  List<Pair> originalLargePairs;

  *box = GetBoundingRectangle(originalPts, 
			      (largePairs)? &originalLargePairs: NULL);
#if DEBUG
  cout << "Preprocess got a point set with bounding box " << *box <<
    endl;
#endif

// MODIFICATION Sat Nov 21 20:39:20
  box->EnlargeWith( 2*epsilon );

  box->MakeIntoSquare();

#if ( ! SIMPLE_MESH ) // this is only needed for the rotated grids
  box->EnlargeByFactor( 1.4142135623730950488016887242097 );
#endif

  SPoint centerPt = box->GetCenter();

  box->SetCenter(0.0, 0.0);

  ChangeCoordinates( originalPts, newPts, centerPt );

  if (largePairs)
    ChangeCoordinates( &originalLargePairs, largePairs, centerPt );

  Transformation t ( centerPt, 0.0 );
  return t;
}


BoundingBox
Algorithm_One::GetBoundingRectangle(List<SPoint> *points, 
				    List<Pair> *pairs) const
{
  assert( points );
  SPoint initP = points->Front();
  BoundingBox box( initP, initP ); // the min and the max are initialized
  List<SPoint> sideCandidates [4], sides [4];

  // initialize all sides with this same point
  if (pairs) {
    sideCandidates[ MinX ].Prepend(initP);
    sideCandidates[ MaxX ].Prepend(initP);
    sideCandidates[ MinY ].Prepend(initP);
    sideCandidates[ MaxY ].Prepend(initP);
  }

  // find out the bounding box
  ListIterator<SPoint> iteratePts( points );
  for (; !iteratePts.IsDone(); iteratePts.Next()) {
    SPoint p = iteratePts.Item();

    if (p.x < box.min.x) {
      box.min.x = p.x;
      // prepend so that the best are in front
      if (pairs) sideCandidates[ MinX ].Prepend(p); 
    } // using 'else if' to save 2 comparisons
    else if (p.x > box.max.x) {
      box.max.x = p.x;
      // prepend so that the best are in front
      if (pairs) sideCandidates[ MaxX ].Prepend(p); 
    }

    if (p.y < box.min.y) {
      box.min.y = p.y;
      // prepend so that the best are in front
      if (pairs) sideCandidates[ MinY ].Prepend(p); 
    } // using 'else if' to save 2 comparisons
    else if (p.y > box.max.y) {
      box.max.y = p.y;
      // prepend so that the best are in front
      if (pairs) sideCandidates[ MaxY ].Prepend(p);
    }
  }
#if DEBUG
  //cout << "in GetBoundingRectangle the box is " << box << endl;
#endif
  // if nobody requested large pairs, just return
  if (pairs == NULL) 
    return box;

  // pick the index of the sides with bigger distance in between
  double lengthX = box.max.x - box.min.x;
  double lengthY = box.max.y - box.min.y;
  int fartherSidesIndex = ( lengthX >= lengthY ) ? MinX : MinY;
    
  // here, we'll use the fact that the best points in the front 
  PickSidePoints( sideCandidates, fartherSidesIndex, sides );

  // from the sides that are the farthest apart, pick points and make pairs
  MakePairs( sides, fartherSidesIndex, pairs );

  return box;
}
  
// here, we'll use the fact that the best points in the front 
// we are picking only the points laying on the two box sides that are the
// futhest away from each other (delta X or delta Y between the min and max)
void
Algorithm_One::PickSidePoints( List<SPoint> temp[], int startIndex,
			       List<SPoint> sides[] ) const
{
  assert( temp && sides );
  assert( startIndex == MinX || startIndex == MinY );

  for ( int i = startIndex; i <= startIndex+1; i++) {
    assert( !temp[i].IsEmpty() && sides[i].IsEmpty() );

    // here, we rely on the fact that the best point is in front
    SPoint firstPt = temp[i].Front();

    ListIterator<SPoint> iteratePts( &temp[i] );
    for ( ; !iteratePts.IsDone(); iteratePts.Next()) {
      SPoint p = iteratePts.Item();
      switch( startIndex ) {
      case MinX: // if we are dealing with the X sides
	if (p.x == firstPt.x) sides[i].Append(p);
	break;
      case MinY: // if we are dealing with the Y sides
	if (p.y == firstPt.y) sides[i].Append(p);
	break;
      default:
	assert( false );
      }
    }
    // we MUST have some points left on the 2 sides we just dealt with
    assert( !sides[i].IsEmpty() );
  }
}

void
Algorithm_One::MakePairs(  List<SPoint> sides[], int startIndex,
			   List<Pair> *pairs ) const
{
  assert( sides && pairs );
  assert( !sides[startIndex].IsEmpty() && !sides[startIndex+1].IsEmpty() );
  assert( startIndex == MinX || startIndex == MinY );

  List<SPoint> *uppperSide = &( sides[ startIndex ] );
  List<SPoint> *lowerSide = &( sides[ startIndex + 1 ] );
  
  ListIterator<SPoint> iterateUpper( uppperSide );
  ListIterator<SPoint> iterateLower( lowerSide );

  for ( ; !iterateUpper.IsDone() && !iterateLower.IsDone(); 
	iterateUpper.Next(), iterateLower.Next()) 
  {
    Pair madePair( iterateUpper.Item(), iterateLower.Item() );
    pairs->Append( madePair );
  }

  // we need at least one pair !!!
  assert( !pairs->IsEmpty() );
}

  


void
Algorithm_One::ChangeCoordinates( List<SPoint> *originalPts, 
				  List<SPoint> *newPts, 
				  SPoint centerPt ) const
{
  assert( originalPts && newPts );
  ListIterator<SPoint> allOriginalPts( originalPts );
  for (; !allOriginalPts.IsDone(); allOriginalPts.Next()) {
    SPoint p = allOriginalPts.Item();
    p = p - centerPt;
    newPts->Append(p);
  }
}  
  
void
Algorithm_One::ChangeCoordinates( List<Pair> *originalPairs, 
				  List<Pair> *newPairs, 
				  SPoint centerPt ) const
{
  assert( originalPairs && newPairs );
  ListIterator<Pair> allOriginalPairs( originalPairs  );
  for (; !allOriginalPairs.IsDone(); allOriginalPairs.Next()) {
    Pair pair = allOriginalPairs.Item();
    pair.p1 -= centerPt;
    pair.p2 -= centerPt;
    newPairs->Append(pair);
  }
}  

  

double 
Algorithm_One::GetDistance(Pair pair) const
{
  return GetDistance(pair.p1, pair.p2);
}

double 
Algorithm_One::GetDistance(SPoint p1, SPoint p2) const
{
  double x = p1.x - p2.x;
  double y = p1.y - p2.y;

  return sqrt(SQR(x) + SQR(y));
}

void
ReadPointsFromFile(const char *name, List<SPoint> *ptList)
{
  int read = 0;
  assert( ptList && name && name[0] );

  FILE *file = fopen(name, "r");
  if (file < 0) {
    cout << "The filename " << file << "was not found... Exiting." << endl;
    exit(-1);
  }
  
  while (1) {
    double x, y;
    read = fscanf( file, "%lf%lf\n", &x, &y );
    if ( read == EOF )
      break;

    assert (read == 2 );
    SPoint p (x,y);
    ptList->Append( p );
  }
  assert( ! ptList->IsEmpty() );

#if DEBUG
  cout << "Read in " << ptList->NumInList() << " points " << endl;
#endif
}

void
ReadPointsFromRscFile(const char *name, List<SPoint> *ptList)
{

    std::ifstream str(name,std::ios::binary);
    if (!str) {
        cout << "The filename " << name << "was not found... Exiting." << endl;
        exit(-1);
    }

    int len;
    str.read((char*)&len,sizeof(int));

    if (len<=0)
        return;

    for (int i=0; i<len; i++)
    {
        double x, y;
        str.read((char*)&(x),sizeof(double));
        str.read((char*)&(y),sizeof(double));
        SPoint p (x,y);
        ptList->Append( p );
    }

    str.close();

    assert( ! ptList->IsEmpty() );

#if DEBUG
    cout << "Read in " << ptList->NumInList() << " points " << endl;
#endif
}

static
void
WritePointsToFile(const char *name, List<SPoint> *ptList)
{
  assert( ptList );
  FILE *file = fopen(name, "w");
  if (file < 0) {
    cout << "The filename could not be open for writing... Exiting." << endl;
    exit(-1);
  }
  
  ListIterator<SPoint> pointIterator( ptList ); 
  for (; !pointIterator.IsDone(); pointIterator.Next()) {
    SPoint thePoint = pointIterator.Item();
    fprintf( file, "%lf %lf\n", thePoint.x, thePoint.y );
  }
}

void
Algorithm_One::SelfTest( int numPointsInModel, int numPointsInImage, 
			 double modelWidth, double modelHeight,
			 double imageWidth, double imageHeight )
{
  // this could be a very nice self test
  // create a bunch of random points, make two lists of them
  // create som more random points, put them in the second list
  // the first list is the model, the second list is the image
  // !!!!!!! struct timeval start = {0,0}, end = {0,0};
  // !!!!!!! struct timezone tzp = {0,0};
  clock_t start = clock();
  clock_t tzp;

  SPoint p;
  srand(time(NULL));

  // first check out some of the Mesh functionality by itself
  Mesh::SelfTest();

  // then check the functionality of a Transformation
  Transformation t;
  t.SelfTest();

  // now proceed with the algorithm self-test
  List<SPoint> *imagePoints = new List<SPoint>;
  List<SPoint> *modelPoints = new List<SPoint>;
  List<FormattedResult> *resultList = NULL;

  cout << "Self-testing with the following values: " << endl;
  cout << "number of points in the model: " << numPointsInModel << endl;
  cout << "number of points in the image: " << numPointsInImage << endl;
  cout << "epsilon: " << epsilon << ", beta: " << beta << endl;
  cout << "AlignGridFactor: " << AlignGridFactor << endl;
  cout << "modelWidth: "<< modelWidth << ", modelHeight: " << modelHeight 
       << endl;
  cout << "imageWidth: " << imageWidth << ", imageHeight: " << imageHeight
       << endl;

  // create a random box
  SPoint rangeFrom ( GetRandomValue(-500, 100), GetRandomValue(-500, 100) );
  SPoint rangeTo ( GetRandomValue(rangeFrom.x + 100, 1000), 
		  GetRandomValue(rangeFrom.y + 100, 1000) );
  SPoint imageRangeFrom = rangeFrom;
  SPoint imageRangeTo = rangeTo;

  if ( modelWidth > 0.0 && modelHeight > 0.0 && 
       imageWidth > 0.0 && imageHeight > 0.0 && 
       modelWidth < imageWidth && modelHeight < imageHeight ) {
    cout << "Using the user supplied values ..." << endl;
    rangeTo.x = rangeFrom.x + modelWidth;
    rangeTo.y = rangeFrom.y + modelHeight;

    imageRangeTo.x = imageRangeFrom.x + imageWidth;
    imageRangeTo.y = imageRangeFrom.y + imageHeight;
  }
  cout << "The sample model box is " << rangeFrom << rangeTo << endl;
  cout << "The sample image box is " << imageRangeFrom << imageRangeTo << endl;
  assert( imageWidth > 0 && imageHeight > 0 && modelWidth > 0 &&
          modelHeight > 0 );
  
  SPoint modelToImageVector = rangeFrom - imageRangeFrom;
  
  // get the random rotation and translation
  SPoint translate ( GetRandomValue(0, imageWidth), 
                    GetRandomValue(0, imageHeight) );
  double rotate =  GetRandomValue(-2*M_PI, 2*M_PI);
  Transformation randomT ( translate, rotate );
  cout << "The random transformation was " << randomT << endl;

  // create two identical lists of points (imagePoints and modelPoints)
  cout << "creating the model in the image (identical points)" << endl; 
  for(int i = 0; i < numPointsInModel; i++) {
    p.x = GetRandomValue(rangeFrom.x, rangeTo.x);
    p.y = GetRandomValue(rangeFrom.y, rangeTo.y);

    // transform the image points with the random transform
    imagePoints->Append( randomT.Transform( p - modelToImageVector ) );
    modelPoints->Append(p);
  }
  
  // continue only with the imagePoints
  cout << "now creating the rest of the image ..." << endl;
  for( i = 0; i < numPointsInImage - numPointsInModel; i++) {
    p.x = GetRandomValue(imageRangeFrom.x, imageRangeTo.x);
    p.y = GetRandomValue(imageRangeFrom.y, imageRangeTo.y);
    //cout << "made a random point #" << i << " : " << p << endl;
    imagePoints->Append(p);
  }

  cout << "Writing the samples to files model and image." << endl;
  WritePointsToFile( "model", modelPoints );
  WritePointsToFile( "image", imagePoints );
  
  cout << "===================== Now matching ..." << endl;
  // !!!!!!! gettimeofday(&start, &tzp);
  resultList = Match(modelPoints, imagePoints);
  //gettimeofday(&end, &tzp);
  tzp = clock();
  cout << "===================== End of Matching ... " << endl;

  clock_t secs = (tzp-start)/CLOCKS_PER_SEC;
  clock_t msecs = (1000.*(tzp-start))/CLOCKS_PER_SEC;
  // !!!!!!! long secs = end.tv_sec - start.tv_sec;
  // !!!!!!! long microsecs = secs*1000000 + (end.tv_usec - start.tv_usec) ;

  cout << "The matching took " << (msecs/(float)1000) 
       << " seconds." << endl;

  if ( ! resultList->IsEmpty() )
    cout << "It's a success !!!" << endl;
  else
    cout << "It was a failure." << endl;
}

  


bool
FindArg(const char *key, const int argc, char** argv, char **line)
{
  for (int i = 1; i < argc; i++) {
    if ( strcmp(key, argv[i]) == 0 ) {
      //      cout << pos << " : " << argv[i+1] << endl;
      return ((*line = argv[i+1]) != NULL);
    }
  }
  return false;
}



static
void
WriteResultsToFile( List<FormattedResult> *results, const char *modelFile )
{
  char newname[100];
  int i = 1;

  assert( results );

  ListIterator<FormattedResult> resultIterator( results ); 
  for (; !resultIterator.IsDone(); resultIterator.Next(), i++) {
    FormattedResult aResult = resultIterator.Item();
    sprintf( newname, "%s.out.%d", modelFile, i );
#if DEBUG
    cout << "Writing result #" << i << " to " << newname << endl;
    cout << "The transform was " << aResult.transform << endl << endl;
#endif
    WritePointsToFile( newname, aResult.transformedModel );
  }
}  


