#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <iostream.h>
//#include "boolfix.h"
#define MAX(x,y) ((x > y)? x : y)
#define SQR(x) ((x)*(x))

#include "list.h"

// =============================  POINT  ========================

class SPoint 
{
 public:

  double x,y;

  SPoint(double _x = 0, double _y = 0): x(_x), y(_y) {};

  bool operator== (const SPoint &otherPt) const
    {return ( (this->x == otherPt.x) && (this->y == otherPt.y) ); } 

  SPoint operator+ (const SPoint &other) const
    {SPoint p( x + other.x, y + other.y); return p; }

  SPoint operator- (const SPoint &other) const
    {SPoint p( x - other.x, y - other.y); return p; }

  SPoint operator+= (const SPoint &other)
    {x += other.x; y += other.y; SPoint p(x,y); return p;}

  SPoint operator-= (const SPoint &other)
    {x -= other.x; y -= other.y; SPoint p(x,y); return p;}

  // why this doesn't work ???????
  ostream& operator<< (ostream& out)
    {return out << "(" << x << ',' << y << ")";}
};


// =============================  INDEX  ========================

class Index
{
 public:
  long x,y;

  Index(long _x = 0, long _y = 0) {x = _x; y = _y;}

  bool operator== (const Index &other) const
    {return ( (this->x == other.x) && (this->y == other.y) ); } 

  bool operator!= (const Index &other) const
    {return ( ! ( (*this) == other ) ); }

  Index operator+ (const Index &other) const
    {Index n( x + other.x, y + other.y); return n; }
};
// =========================== BOUNDING BOX  ========================

class BoundingBox
{
 public:
  SPoint min, max;
  BoundingBox() {};
  BoundingBox(SPoint _min, SPoint _max) : min(_min), max(_max) {}
  void MakeIntoSquare( void );
  void EnlargeByFactor( double factor );
  void EnlargeWith( double delta );
  SPoint GetCenter( void ) const;
  SPoint SetCenter( double x, double y );
};


// =============================  PAIR  ========================

class Pair
{
 public:
  SPoint p1, p2;

  Pair() {};
  inline Pair(SPoint _p1, SPoint _p2): p1(_p1), p2(_p2) {};

  Pair Reverse() { Pair pair (p2, p1); return pair; }

  bool operator== (const Pair &theOtherPair) {
    return ((p1 == theOtherPair.p1 && p2 == theOtherPair.p2 ) ||
      (p1 == theOtherPair.p2 && p2 == theOtherPair.p1));
  }

  Pair operator+= (const Pair &other)
    {p1 += other.p1; p2 += other.p2; Pair pair(p1,p2); return pair;}

  Pair operator- (const Pair &other)
    {Pair pair (p1 - other.p1, p2 - other.p2); return pair; }

};

// ==========================  TRANSFORMATION  ========================

class Transformation
{
 public:
  SPoint translateV;
  double theta;

  Transformation() { theta = 0.0; } // the point should be init by default
  Transformation(SPoint _deltaV, double _theta) {Init(_deltaV, _theta);}
  Transformation(Pair modelPair, Pair imagePair) 
    { GetTransform( modelPair, imagePair); }
  
  inline void Init(SPoint _deltaV, double _theta) {
      translateV = _deltaV;
      theta = _theta;
  }

  // We rotate p1,p2 so that it's parallel to q1,q2 and then translate 
  // it. This transformation assumes rotation about (0,0), so 
  // everything works out..
  void GetTransform ( Pair modelPair, Pair imagePair);
  double GetAngleWithXaxis(Pair pair) const;

  Transformation Compose( Transformation &other );
  inline Pair Transform ( Pair pair ) const;
  inline SPoint Transform ( SPoint pt ) const;
  inline List<SPoint>* Transform( List<SPoint> *list ) const;
  

  bool operator== (const Transformation &other) {
    return ( translateV == other.translateV && theta == other.theta );
  }

  static inline SPoint GetPoint(SPoint center, double radius, double angle);

  static inline SPoint Rotate( SPoint p, double angle );
  static inline Pair Rotate( Pair pair, double angle );
  static inline BoundingBox Rotate( BoundingBox box, double angle );

  static inline bool IsLengthBetweenSQR(double length, double from, double to);
  static inline bool IsLengthBetweenSQR(Pair pair, double from, double to);
  static inline bool IsLengthBetweenSQR(SPoint p1, SPoint p2, double from, double to);

  static inline double LengthSQR( SPoint p1, SPoint p2 );

 public:
  void SelfTest();
};

// ======================  FORMATTED RESULT  ========================

class FormattedResult
{
 public:
  FormattedResult() {transformedModel = NULL;}
  FormattedResult(Transformation t, List<SPoint> *modelPoints = NULL) 
    { transformedModel = modelPoints; transform = t;}

  // needed for the list stuff
  bool operator== (FormattedResult &other) 
    {return (transformedModel == other.transformedModel && 
	     transform == other.transform);}

  List<SPoint> *transformedModel;
  Transformation transform;
  
};


// =============================  GRID  ========================

class Grid
{
 public:
  Grid(double _angle, BoundingBox box, 
       double _cellSizeOnX, double _cellSizeOnY);
  ~Grid();

  void InitializeWith( List<SPoint> *points );
  void InsertPoint(SPoint p);

  void Get4Buckets(SPoint samplePt, List< List<SPoint>* > *buckets) const;
  void GetFBuckets(SPoint samplePt, List< List<SPoint>* > *buckets) const;
  long GetElements() { return elementsOnX; }
#if ( SIMPLE_MESH && ! STAMP_ANNULUS )
  void GetBucketsInAnulus(SPoint centerPoint, double radius, double epsilon,
			  List< List<SPoint>* > *allBuckets ) const;
  void GetBucketsInAnulusFast( SPoint centerPoint, 
			       double radius, double epsilon,
			       List< List<SPoint>* > *allBuckets ) const;
  void GetBucketsInAnulusLinearly (SPoint centerPt, double radius, 
				   double epsilon,
				   List< List<SPoint>* > *allBuckets ) const;

#elif ( SIMPLE_MESH && STAMP_ANNULUS )
 public:
  void PreStampAnnulus( double radius, double epsilon );
 private:
  void StoreSqrAnnulus( SPoint centerPoint, 
			double radius, double epsilon );
  void StoreLinAnnulusSquarely ( SPoint centerPoint, 
				 double radius, double epsilon );
  void StoreLinAnnulusLinearly ( SPoint centerPoint, 
				 double radius, double epsilon );

 public:
  void GetPreStampedBuckets( const SPoint &centerPoint, 
				    List<List<SPoint>*> *allBuckets );

  
#endif


 private:

  List<SPoint>* GetBucket(SPoint samplePt) const;
  Index inline GetIndex(SPoint p) const;
  SPoint GetPositionWithinBucket(SPoint p) const;
  bool inline PointIsOK(SPoint p) const;
  bool inline IndexIsOK(Index n) const;
  double inline ClosestWall( double norm, double smaller, double bigger ) const;
  double inline FarthestWall( double norm, double smaller, double bigger ) const;
  double inline ClosestBetweenWall( double norm, double smaller, double bigger ) const;
  double inline FarthestBetweenWall( double norm, double smaller, double bigger ) const;
  SPoint PointInGrid( SPoint p ) const;
  bool inline IsBucketInAnnulus( Index n, SPoint centerPoint, 
				 double radius, double epsilon ) const;
  void DisplayIndexCenter( SPoint min, Index n, int q, int t );

  // the grid is going to be implemented with a dynamically allocated
  // 2-dimensional array of List pointers
  List<SPoint> ***grid;
  
  double cellSizeOnX, cellSizeOnY;
  double areaSizeOnX, areaSizeOnY;
  BoundingBox ptBox;  // the min and max points -- for sanity checks
  long elementsOnX, elementsOnY;
  double angle;
  
#if ( SIMPLE_MESH && STAMP_ANNULUS )
  List<Index> bucketsInAnnulus;
#endif
};

// Notice that we might (and should if the math was correct)
// get redundant grids -- one for rotationAngle alpha, 
// the other one for alpha+pi
// ??? We should probably just make then sqrt(r)/2 grids 
// and handle internally any rotationalAngles > pi as (pi - alpha).

// =============================  MESH  ========================

class Mesh
{
 public:
  Mesh(BoundingBox box, double radius, double epsilon, double beta,
       List<SPoint> *imagePoints);
  ~Mesh();


  void FindPossiblePairEnds(SPoint &centerPoint, 
			    List< List<SPoint>* > *allBuckets) const;

 private:
  void InitializeWith(List<SPoint> *allImagePoints);

 private:
  // the mesh is implemented as an array of grids indexed
  // on the rotaionalAngle value
  double cellSizeOnX, cellSizeOnY;
  double areaSizeOnX, areaSizeOnY;
  BoundingBox box;
  double epsilon, beta;
  double radius;
  Grid **grids;
  double rotationalStep; 
  int numberOfSteps;
  SPoint *samplePoints;

 public:
  static void SelfTest();
};


// =========================  SIMPLE MESH  ========================

class SimpleMesh
{
 public:
  SimpleMesh(BoundingBox box, double radius, double epsilon, double beta,
	     List<SPoint> *imagePoints);
  ~SimpleMesh();


  void FindPossiblePairEnds(SPoint centerPoint, 
			    List< List<SPoint>* > *allBuckets) const;

 private:

  double cellSizeOnX, cellSizeOnY;
  double areaSizeOnX, areaSizeOnY;
  BoundingBox box;
  double epsilon, beta;
  double radius;
  Grid *grid;
};


// =========================  ALIGNMENT GRID  ========================

class AlignmentGrid
{
 public:
  AlignmentGrid(BoundingBox box, double epsilon, List<SPoint> *imagePoints);
  ~AlignmentGrid();

  bool Align( List<SPoint> *modelPoints, Transformation t ) const;

 private:
  BoundingBox box;
  double epsilon;

  // these vars are kept for convenience
  double cellSizeOnX, cellSizeOnY;
  double areaSizeOnX, areaSizeOnY;


  // the alignment grid is simply a grid
  Grid *grid;

  bool HasAtLeastOneMatch( SPoint thePoint, 
			   List< List<SPoint>* > *buckets ) const;

 public:
  void SelfTest();
};

// =========================  ALIGNMENT BIT GRID  ========================

class AlignmentBitGrid
{
 public:
  AlignmentBitGrid(BoundingBox box, double cellSize, List<SPoint> *imagePoints);
  ~AlignmentBitGrid();

  bool Align( List<SPoint> *modelPoints, Transformation t ) const;

 private:
  enum { Not_Marked = 0, Marked = 1 };
  double epsilon;
  BoundingBox box;
  long elementsOnX, elementsOnY;

  double cellSizeOnX, cellSizeOnY;
  double areaSizeOnX, areaSizeOnY;

  // the alignment grid is implemented as a 2-D array of chars
  char **grid;


  void InitializeWith(List<SPoint> *imagePoints);
  void SpreadPointInGrid( SPoint p );
  double inline ClosestWall( double norm, double smaller, double bigger ) const;
  Index inline GetIndex(SPoint p) const;
  bool PointIsOK(SPoint p) const;
  bool IndexIsOK(Index n) const;
};


// ==========================  ALGORITHM_1  ========================

class Algorithm_One
{
 public:
  static double DefaultEpsilon;
  static double DefaultBeta;
  // #define DefaultEpsilon 1.0
  // #define DefaultBeta 1.0

 private:
  double areaSizeX, areaSizeY;
  double epsilon, beta;
  BoundingBox boxModel, boxImage;
  List<SPoint> pointsInModel, pointsInImage;
  List<Pair> largeModelPairs;

 public:

  Algorithm_One(double _epsilon, double _beta);
  ~Algorithm_One();

  List<FormattedResult>* Match(List<SPoint> *pointsInModel, 
			       List<SPoint> *pointsInImage);

  void SelfTest( int numPointsInModel, int numPointsInImage, 
		 double modelWidth, double modelHeight,
		 double imageWidth, double imageHeight );
  
 private:
  enum { MinX = 0, MaxX, MinY, MaxY };

  // Do some work on the point sets that will put them in a format with
  // which we can work. Notably,
  //    * find their bounding rectangle, 
  //    * pick all of the points that are on its 4 sides (needed later), 
  //    * enlarge the bounding rectangle into a square, 
  //    * enlarge the square by a factor of sqrt(2), 
  //    * find the center of that square, set it to coordinates x=0, y=0
  //    * change the coordinates of all points
  //    * change the coordinates of the points separated into 4 groups
  //    * create a list of pairs made up from the points in these 4 groups
  //    * return the new points, the bounding box, and the pairs
  //    * the return value contains the vector that was used to change the
  //        coordinate system of the points
  // Yes, it's quite a lot, isn't it ?
  Transformation PreprocessPoints(List<SPoint> *originalPts, 
				  List<SPoint> *newPts,
				  BoundingBox *box,
				  List<Pair> *largePairs = NULL) const;
  
  BoundingBox GetBoundingRectangle(List<SPoint> *points, 
				   List<Pair> *pairs) const;

  void PickSidePoints( List<SPoint> temp[], int startIndex,
		       List<SPoint> sides[] ) const;

  void MakePairs( List<SPoint> sides[], int startIndex,
		  List<Pair> *pairs ) const;

  void ChangeCoordinates( List<SPoint> *orginalPts, 
			  List<SPoint> *newPts, 
			  SPoint centerPt ) const;
  
  void ChangeCoordinates( List<Pair> *originalPairs, 
			  List<Pair> *newPairs, 
			  SPoint centerPt ) const;
  
  double GetDistance(Pair pair) const;
  double GetDistance(SPoint p1, SPoint p2) const;

  Transformation GetAllTransformations(Pair pairInModel, 
				       Pair pairInImage) const;

  void GetOneTransformation(Pair modelPair, Pair imagePair, 
			    SPoint *translate, double *rotate) const;

  double GetAngleWithXaxis(Pair pair) const;

  Pair Rotate( Pair pair, double angle ) const;
  SPoint Rotate( SPoint p, double angle ) const;

  bool Align(List<SPoint> *pointsInModel, List<SPoint> *pointsInImage, 
	      Transformation transform) const;


};
  
static 
double 
GetRandomValue(double from, double to) 
{
  double val = ((double) rand())/((double)RAND_MAX);
  assert(val >= 0 && val <= 1);
  double range = to - from;
  assert(range > 0);
  val *= range;
  val += from;
  return val;
}


bool
FindArg(const char *key, const int argc, char** argv, char **line);

static double AlignGridFactor = 1.4142135623730950488016887242097 / Algorithm_One::DefaultBeta;
static double MeshFactor = 2;

void
ReadPointsFromFile(const char *name, List<SPoint> *ptList);

void
ReadPointsFromRscFile(const char *name, List<SPoint> *ptList);


