#include <iostream>
#include <fstream>
#include <direct.h>
#include <ctime>
#include <crtdbg.h>

#include <ReSCue/Point2.h>
#include <ReSCue/polygonal_curve.h>
#include <ReSCue/sampling_matcher.h>
#include <ReSCue/circular_arc_edge.h>
#include <ReSCue/free_space_graph.h>
#include <ReSCue/extended_fs_graph.h>
#include <ReSCue/directed_hausdorff_graph.h>
#include <ReSCue/partial_match_graph.h>
#include <ReSCue/partial_sampling_matcher.h>

#include "testseries.h"

#include "match.h"

using namespace rescue;

typedef Point2<double> RPoint;
typedef polygonal_curve<RPoint> Curve;

typedef sampling_matcher<
            free_space_graph<
                circular_arc_edge<int,RPoint> 
            > 
        > 
        SamplingMatcher;
typedef SamplingMatcher::curve Curve;
typedef SamplingMatcher::vertex Vertex;
typedef sampling_matcher <
            extended_fs_graph <
                circular_arc_edge<int,RPoint> 
            > 
        > 
        FastSamplingMatcher;

typedef partial_sampling_matcher <
            partial_match_graph <
                circular_arc_edge<int,RPoint> 
            > 
        > 
        PartialSamplingMatcher;

typedef partial_sampling_matcher <
            directed_hausdorff_graph <
                circular_arc_edge<int,RPoint> 
            > 
        > 
        HausdorffSamplingMatcher;


// read one non-comment-token
void get_next_token(std::ifstream& ifs, char* buf, int max_length, char delim='\n')
{
    do {
        if (ifs.eof())
            break;
        ifs.getline(buf,max_length,delim);
    } while((strlen(buf)==0 || buf[0]=='%'));
}

void my_testseries(const char* ts_filename)
{
    char out_file[128];
    strcpy(out_file,ts_filename);
    out_file[strlen(out_file)-1] = 'r';
    std::ofstream ofs(out_file);

    char next_token[128];

    std::ifstream ifs;
    ifs.open(ts_filename);

    // Is there another test series to be performed?
    get_next_token(ifs,next_token,128);
    while (!strcmp(next_token,"series"))
    {
        char dir_name[128];
        // Read directory to work with
        get_next_token(ifs,dir_name,128);
        chdir(dir_name);

        // Read Matching method from the file
        char method[32];
        get_next_token(ifs,method,32);
        ofs<<"\n%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n"<<method
           <<"\n%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n"<<"\n";

        // Read Number of iterations
        char numbuf3[32];
        int NUMBER_OF_ITERATIONS = 1;
        get_next_token(ifs,numbuf3,32);
        NUMBER_OF_ITERATIONS  = atoi(numbuf3);

        // Read eps,delta
        get_next_token(ifs,next_token,32,',');
        double eps = (double)atoi(next_token);
        get_next_token(ifs,next_token,32);
        double delta = (double)atoi(next_token);
        ofs<<"eps = "<<eps<<"; "<<"delta = "<<delta<<"\n\n";

        get_next_token(ifs,next_token,128,',');
        
        double duration_sum = 0.0;
        int num_of_curves = 0;

        while (!ifs.eof() && strcmp(next_token,"series") && strcmp(next_token,"end"))
        {        
            char P_name[128];
            char Q_name[128];
            // Read filename pairs (P Q) from the file specified by s
            strcpy(P_name,dir_name);
            strcat(P_name,"\\");
            strcat(P_name,next_token);
            get_next_token(ifs,next_token,128);
            Curve P;
            std::fstream ifs_P(P_name,std::ios::in | std::ios::binary);
            P.read(ifs_P);
            ifs_P.close();
            strcpy(Q_name,dir_name);
            strcat(Q_name,"\\");
            strcat(Q_name,next_token);
            Curve Q;
            std::fstream ifs_Q(Q_name,std::ios::in | std::ios::binary);
            Q.read(ifs_Q);
            ifs_Q.close();

            // Read number of points to be read from P and Q
            //get_next_token(ifs,next_token,128,',');
            //int m=atoi(next_token);
            //get_next_token(ifs,next_token,128);
            //int n=atoi(next_token);
            get_next_token(ifs,next_token,128,',');

            num_of_curves++;
            
            bool match_found;

            double duration;
            std::cout<<P_name<<","<<Q_name<<"\n";

            
            if (!strcmp(method,"sampling_match"))
            {
                
                clock_t start = clock();

                for (int ii=0; ii<NUMBER_OF_ITERATIONS; ii++)
                {
                    SamplingMatcher M(P,Q,eps,delta);
                    match_found = M.match();
                    M.clear();
                }
                clock_t finish = clock();
                duration = 1000.0*(((double)(finish-start))/CLOCKS_PER_SEC)/(double)NUMBER_OF_ITERATIONS;
            
            } else if (!strcmp(method,"sampling_match_discrete")) {
                
                clock_t start = clock();
                for (int ii=0; ii<NUMBER_OF_ITERATIONS; ii++)
                {
                    SamplingMatcher M(P,Q,eps);
                    match_found = M.match();
                    M.clear();
                }
                clock_t finish = clock();
                duration = 1000.0*(((double)(finish-start))/CLOCKS_PER_SEC)/(double)NUMBER_OF_ITERATIONS;
            
            } else if (!strcmp(method,"fast_sampling_match")) {
                
                clock_t start = clock();
                for (int ii=0; ii<NUMBER_OF_ITERATIONS; ii++)
                {
                    FastSamplingMatcher M(P,Q,eps,delta);
                    match_found = M.match();
                    M.clear();
                }
                clock_t finish = clock();
                duration = 1000.0*(((double)(finish-start))/CLOCKS_PER_SEC)/(double)NUMBER_OF_ITERATIONS;
            
            } else if (!strcmp(method,"fast_sampling_match_discrete")) {
                
                clock_t start = clock();
                for (int ii=0; ii<NUMBER_OF_ITERATIONS; ii++)
                {
                    FastSamplingMatcher M(P,Q,eps);
                    match_found = M.match();
                    M.clear();
                }
                clock_t finish = clock();
                duration = 1000.0*(((double)(finish-start))/CLOCKS_PER_SEC)/(double)NUMBER_OF_ITERATIONS;
            
            } else if (!strcmp(method,"partial_match")) {
                
                clock_t start = clock();
                for (int ii=0; ii<NUMBER_OF_ITERATIONS; ii++)
                {            
                    PartialSamplingMatcher M(P,Q,eps,delta);
                    match_found = M.match();
                    M.clear();
                }
                clock_t finish = clock();
                duration = 1000.0*(((double)(finish-start))/CLOCKS_PER_SEC)/(double)NUMBER_OF_ITERATIONS;

            } else if (!strcmp(method,"partial_match_discrete")) {
                
                clock_t start = clock();
                for (int ii=0; ii<NUMBER_OF_ITERATIONS; ii++)
                {
                    PartialSamplingMatcher M(P,Q,eps);
                    match_found = M.match();
                    M.clear();
                }
                clock_t finish = clock();
                duration = 1000.0*(((double)(finish-start))/CLOCKS_PER_SEC)/(double)NUMBER_OF_ITERATIONS;
            
            } else if (!strcmp(method,"directed_hausdorff_match")) {

                if (P.length()>0 && Q.length()>0)
                {
                    clock_t start = clock();
                    for (int ii=0; ii<NUMBER_OF_ITERATIONS; ii++)
                    {
                        HausdorffSamplingMatcher M(P,Q,eps);
                        match_found = M.match();
                        M.clear();
                    }
                    clock_t finish = clock();
                    duration = 1000.0*(((double)(finish-start))/CLOCKS_PER_SEC)/(double)NUMBER_OF_ITERATIONS;
                } else {
                    duration = 0.0;
                    num_of_curves--;
                }

            } else if (!strcmp(method,"hausdorff_GRID")) {

                List<SPoint> model, image;
                List<FormattedResult> *results;
                
                bool empty = true;
                bool empty_p = true;
                bool empty_q = true;

                Curve::vertex_iterator v_it = P.vertex_begin();
                for (; v_it!=P.vertex_end(); v_it++)    
                {
                    empty_p = false;
                    double x = (*v_it).pos.x;
                    double y = (*v_it).pos.y;
                    SPoint p(x,y);
                    model.Append( p );
                }
                v_it = Q.vertex_begin();
                for (; v_it!=Q.vertex_end(); v_it++)    
                {
                    empty_q = false;
                    double x = (*v_it).pos.x;
                    double y = (*v_it).pos.y;
                    SPoint p(x,y);
                    image.Append( p );
                }
                empty = empty_p || empty_q;

                if (!empty)
                {
                    clock_t start = clock();
                    for (int ii=0; ii<NUMBER_OF_ITERATIONS; ii++)
                    {
                        Algorithm_One algo(eps, delta);
                        results = algo.Match( &model, &image );
                        if (results->NumInList()>0)
                            match_found = true;
                    } 

                    clock_t finish = clock();
                    duration = 1000.0*(((double)(finish-start))/CLOCKS_PER_SEC)/(double)NUMBER_OF_ITERATIONS;
                } else {
                    num_of_curves--;
                    duration = 0.0;
                }

            }
            duration_sum += duration;
            ofs<<P_name<<","<<Q_name<<"\n"<<duration<<" msec\n";
            ofs<<"m = "<<P.length()<<"; n = "<<Q.length()<<"\n";
            ofs<<"result = "<<match_found<<"\n\n";

        }
        ofs<<"curve matches performed: "<<num_of_curves<<"\n";
        ofs<<"total duration: "<<duration_sum<<"\n";
        ofs<<"avg. duration: "<<duration_sum/(double)num_of_curves<<"\n";
    }
    ofs.close();

}

