#ifndef RIGIDMOTIONINTERVAL_H
#define RIGIDMOTIONINTERVAL_H

#include "angle.h"
#include "rigid_motion.h"
#include "rescuenumerics.h"
#include "rescutil.h"
#include <iostream>

namespace rescue
{

/** This class is part of the implementation of an exponential-time-algorithm for matching 
with respect to the Frechet-distance, provided in the class \c covering_matcher. 
This implementation exists only for historical reasons (implemented as a "proof of concept" 
of intersecting projected transporter subsets) and is thus left undocumented in wide parts.
*/    

// ==========================================================================================
// ==========================================================================================

template <class arithmetic>
class rigid_motion_interval
{

private:
    

public:

    typedef unit_circle_angle<arithmetic> Angle;

    bool empty;
    
    bool all_angles;

    arithmetic min_t_x;
    arithmetic min_t_y;
    arithmetic min_sin_phi;
    arithmetic min_cos_phi;

    arithmetic max_t_x;
    arithmetic max_t_y;
    arithmetic max_sin_phi;
    arithmetic max_cos_phi;

    rigid_motion_interval();

    // binrer Operator zur Schnittmengenberechnung zweier rigid_motion_intervals
    rigid_motion_interval operator & (const rigid_motion_interval<arithmetic>& ri);

    operator rigid_motion<arithmetic> ();

    friend std::ostream& operator<<(std::ostream& o, const rigid_motion_interval<arithmetic> I);

    inline bool EMPTY() {return empty;}


};


// ==========================================================================================
// ==========================================================================================



// ==========================================================================================
template <class arithmetic>
rigid_motion_interval<arithmetic> 
    rigid_motion_interval<arithmetic>::operator & (
    const rigid_motion_interval<arithmetic>& ri)
// ==========================================================================================
{
    #ifdef DEBUG_INTERVAL_INTERSECTIONS
    std::cerr<<"intersecting:\n"<<*this<<ri<<"result:\n";
    #endif

    rigid_motion_interval<arithmetic> ret_i;
    ret_i.empty = true;
    ret_i.all_angles = all_angles && ri.all_angles;

    if (empty || ri.empty)
    {
        #ifdef DEBUG_INTERVAL_INTERSECTIONS
        std::cerr<<ret_i;
        #endif
        return ret_i;
    }
    
    // Zuerst Durschnitt der Translationen berechnen...

    // x-Komponente...
    if (max_t_x<ri.min_t_x || min_t_x>ri.max_t_x)
    {
        // Durchschnitt leer!
        #ifdef DEBUG_INTERVAL_INTERSECTIONS
        std::cerr<<ret_i;
        #endif
        return ret_i;
    }
    if (min_t_x>=ri.min_t_x && min_t_x<=ri.max_t_x)
    {
        ret_i.min_t_x = min_t_x;
    } else {
        ret_i.min_t_x = ri.min_t_x;
    }
    if (max_t_x>=ri.min_t_x && max_t_x<=ri.max_t_x)
    {
        ret_i.max_t_x = max_t_x;
    } else {
        ret_i.max_t_x = ri.max_t_x;
    }

    // y-Komponente...
    if (max_t_y<ri.min_t_y || min_t_y>ri.max_t_y)
    {
        // Durchschnitt leer!
        #ifdef DEBUG_INTERVAL_INTERSECTIONS
        std::cerr<<ret_i;
        #endif
        return ret_i;
    }
    if (min_t_y>=ri.min_t_y && min_t_y<=ri.max_t_y)
    {
        ret_i.min_t_y = min_t_y;
    } else {
        ret_i.min_t_y = ri.min_t_y;
    }
    if (max_t_y>=ri.min_t_y && max_t_y<=ri.max_t_y)
    {
        ret_i.max_t_y = max_t_y;
    } else {
        ret_i.max_t_y = ri.max_t_y;
    }

    if (all_angles)
    {
        ret_i.min_sin_phi = ri.min_sin_phi;
        ret_i.min_cos_phi = ri.min_cos_phi;
        ret_i.max_sin_phi = ri.max_sin_phi;
        ret_i.max_cos_phi = ri.max_cos_phi;
        ret_i.empty = false;
        #ifdef DEBUG_INTERVAL_INTERSECTIONS
        std::cerr<<ret_i;
        #endif
        return ret_i;
    } else if (ri.all_angles) {
        ret_i.min_sin_phi = min_sin_phi;
        ret_i.min_cos_phi = min_cos_phi;
        ret_i.max_sin_phi = max_sin_phi;
        ret_i.max_cos_phi = max_cos_phi;
        ret_i.empty = false;
        #ifdef DEBUG_INTERVAL_INTERSECTIONS
        std::cerr<<ret_i;
        #endif
        return ret_i;
    }

    // Nun zum Durchschnitt der Rotation...
    if (Angle(ri.min_sin_phi,ri.min_cos_phi).between(Angle(min_sin_phi,min_cos_phi),Angle(max_sin_phi,max_cos_phi)))
        if (Angle(ri.max_sin_phi,ri.max_cos_phi).between(Angle(min_sin_phi,min_cos_phi),Angle(max_sin_phi,max_cos_phi)))
        {
            ret_i.min_sin_phi = ri.min_sin_phi;
            ret_i.min_cos_phi = ri.min_cos_phi;
            ret_i.max_sin_phi = ri.max_sin_phi;
            ret_i.max_cos_phi = ri.max_cos_phi;
            ret_i.empty = false;
            #ifdef DEBUG_INTERVAL_INTERSECTIONS
            std::cerr<<ret_i;
            #endif
            return ret_i;
        } else {
            ret_i.min_sin_phi = ri.min_sin_phi;
            ret_i.min_cos_phi = ri.min_cos_phi;
            ret_i.max_sin_phi = max_sin_phi;
            ret_i.max_cos_phi = max_cos_phi;
            ret_i.empty = false;
            #ifdef DEBUG_INTERVAL_INTERSECTIONS
            std::cerr<<ret_i;
            #endif
            return ret_i;
        }
    if (Angle(min_sin_phi,min_cos_phi).between(Angle(ri.min_sin_phi,ri.min_cos_phi),Angle(ri.max_sin_phi,ri.max_cos_phi)))
        if (Angle(max_sin_phi,max_cos_phi).between(Angle(ri.min_sin_phi,ri.min_cos_phi),Angle(ri.max_sin_phi,ri.max_cos_phi)))
        {
            ret_i.min_sin_phi = min_sin_phi;
            ret_i.min_cos_phi = min_cos_phi;
            ret_i.max_sin_phi = max_sin_phi;
            ret_i.max_cos_phi = max_cos_phi;
            ret_i.empty = false;
            #ifdef DEBUG_INTERVAL_INTERSECTIONS
            std::cerr<<ret_i;
            #endif
            return ret_i;
        } else {
            ret_i.min_sin_phi = min_sin_phi;
            ret_i.min_cos_phi = min_cos_phi;
            ret_i.max_sin_phi = ri.max_sin_phi;
            ret_i.max_cos_phi = ri.max_cos_phi;
            ret_i.empty = false;
            #ifdef DEBUG_INTERVAL_INTERSECTIONS
            std::cerr<<ret_i;
            #endif
            return ret_i;
        }

    #ifdef DEBUG_INTERVAL_INTERSECTIONS
    std::cerr<<ret_i;
    #endif
    return ret_i;

}


// ==========================================================================================
template<class arithmetic>
std::ostream& operator<<(std::ostream& o, const rigid_motion_interval<arithmetic> I)
// ==========================================================================================
{
    if (I.empty)
    {
        o<<"[empty]\n";
        return o;
    }
    o<<"(";
    // Ausgabe ist das Minimum...
    //if (!I.all_angles && I.min_sin_phi<=0.0 && I.min_cos_phi<=0.0)
    //    o<<"-";
    double d_phi = (atan(I.min_sin_phi/I.min_cos_phi));
    if (I.min_sin_phi<0.0 && I.min_cos_phi>0.0)
        d_phi += 2.0*PI;
    else if (I.min_sin_phi>0.0 && I.min_cos_phi<0.0)
        d_phi += PI;
    else if (I.min_sin_phi<0.0 && I.min_cos_phi<0.0)
        d_phi += PI;
    double d_x = (I.min_t_x);
    double d_y = (I.min_t_y);
    if (!I.all_angles)
        o<<d_phi<<","<<d_x<<","<<d_y<<")";
    else
        o<<"*"<<","<<d_x<<","<<d_y<<")";

    o<<"...(";
    d_phi = (atan(I.max_sin_phi/I.max_cos_phi));
    if (I.max_sin_phi<0.0 && I.max_cos_phi>0.0)
        d_phi += 2.0*PI;
    else if (I.max_sin_phi>0.0 && I.max_cos_phi<0.0)
        d_phi += PI;
    else if (I.max_sin_phi<0.0 && I.max_cos_phi<0.0)
        d_phi += PI;
    d_x = (I.max_t_x);
    d_y = (I.max_t_y);
    if (!I.all_angles)
        o<<d_phi<<","<<d_x<<","<<d_y<<")"<<std::endl;
    else
        o<<"*"<<","<<d_x<<","<<d_y<<")"<<std::endl;
    return o;
}

// ==========================================================================================
template<class arithmetic>
rigid_motion_interval<arithmetic>::rigid_motion_interval()
// ==========================================================================================
{
    empty = false;
}


// ==========================================================================================
template<class arithmetic>
rigid_motion_interval<arithmetic>::operator rigid_motion<arithmetic> ()
// ==========================================================================================
{
    rigid_motion<arithmetic> r;
    r.t_x = (min_t_x + max_t_x) / 2.0;
    r.t_y = (min_t_y + max_t_y) / 2.0;

    return r;

}

}

#endif
