
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <fstream>
#include <iostream>
#include <stdio.h>
#include <io.h>
#include <math.h>
#include <float.h>
#include <set>
#include <list>

int len;
double* xx;
double* yy;

// bounding box information
double bb_min_x;
double bb_max_x;
double bb_min_y;
double bb_max_y;

#define PI 3.1415926535897932384626433832795

void read_squid(std::ifstream& str)
// ==========================================================================================
{

    //clear();

    if (xx!=NULL)
        delete[] xx;
    if (yy!=NULL)
        delete[] yy;
    len =  0;

    char  ch;
    str.read(&ch,sizeof(char));
    if (ch != '#')
        return;

    str>>len;

    xx = new double[len];
    yy = new double[len];

    for (int i=0; i<len; i++)
    {
        int x,y;
        str>>x;
        str>>y;
        xx[i] = 2.0*(double)x;
        yy[i] = 2.0*(double)y;

        //std::cout<<xx[i]<<", "<<yy[i]<<std::endl;
        //point p((double)xx,(double)yy);
        //push_back(new vertex(p));
    }

}

void write_rsc(std::ofstream& str, int my_len)
// ==========================================================================================
{

    if (len<=0)
        return;

    str.write((char*)&my_len,sizeof(int));

    for (int i=0; i<len; i++)
    {
        str.write((char*)&(xx[i]),sizeof(double));
        str.write((char*)&(yy[i]),sizeof(double));
    }

}


void read_rsc(std::ifstream& str)
// ==========================================================================================
{

    str.read((char*)&len,sizeof(int));

    // determine bounding_box
    bb_min_x = DBL_MAX;
    bb_min_y = DBL_MAX;
    bb_max_x = DBL_MIN;
    bb_max_y = DBL_MIN;

    if (len<=0)
        return;

    if (xx!=NULL)
        delete[] xx;
    if (yy!=NULL)
        delete[] yy;

    xx = new double[len];
    yy = new double[len];

    for (int i=0; i<len; i++)
    {
        str.read((char*)&(xx[i]),sizeof(double));
        str.read((char*)&(yy[i]),sizeof(double));
        if (xx[i]<bb_min_x)
            bb_min_x = xx[i];
        if (xx[i]>bb_max_x)
            bb_max_x = xx[i];
        if (yy[i]<bb_min_y)
            bb_min_y = yy[i];
        if (yy[i]>bb_max_y)
            bb_max_y = yy[i];
    }

    int ii=5;
    return;

}

double transform_x(double px, double py, double angle, double t_x, double t_y)
{
    double s = sin(angle);
    double c = cos(angle);
    return s*px+c*py+t_x;
}

double transform_y(double px, double py, double angle, double t_x, double t_y)
{
    double s = sin(angle);
    double c = cos(angle);
    return c*px-s*py+t_y;
}


void write_rsc_query(std::ofstream& str, double box_l_x, double box_l_y, double box_u_x, double box_u_y)
// ==========================================================================================
{
 
    if (len<=0)
        return;

    std::list<int> index_list;

    for (int i=0; i<len; i++)
    {
        if (   xx[i]>=box_l_x && xx[i]<=box_u_x 
            && yy[i]>=box_l_y && yy[i]<=box_u_y )
        {
            index_list.push_back(i);
        }
    }

 	int my_len = index_list.size();
    str.write((char*)&my_len,sizeof(int));
    std::list<int>::iterator it = index_list.begin();
    for (; it!=index_list.end(); it++)
    {
        double tx = transform_x(xx[*it],yy[*it],7.0*PI/24.0,200.0,200.0);
        double ty = transform_y(xx[*it],yy[*it],7.0*PI/24.0,200.0,200.0);
        str.write((char*)&(tx),sizeof(double));
        str.write((char*)&(ty),sizeof(double));
    }

}

void write_rsc_pattern(std::ofstream& str, int max_len)
// ==========================================================================================
{

    if (len<0 || max_len>len)
        return;

    std::set<int> index_set;
    std::list<int> pattern_list;

    for (int i=0; i<len; i++)
        index_set.insert(i);
    for (i=0; i<max_len; i++)
    {
        // pick an index from index_set at random
        std::set<int>::iterator it = index_set.begin();
        int sz = index_set.size();
        int rnd = rand();
        int rnd_max = RAND_MAX;
        int rnd_idx = (sz*rnd)/(rnd_max+1);
        //it += rnd;
        for (int j=0; j<rnd_idx; j++)
            it++;
        int idx = *it;
        if (idx<0)
            std::cout<<"??\n";
        pattern_list.push_back(*(it));
        index_set.erase(*(it));
    }



	int my_len = pattern_list.size();
    std::list<int>::iterator it = pattern_list.begin();

    str.write((char*)&my_len,sizeof(int));

    for (; it!=pattern_list.end(); it++)
    {
        int idx = *it;
        str.write((char*)&(xx[idx]),sizeof(double));
        str.write((char*)&(yy[idx]),sizeof(double));
        if (!str)
            std::cout<<"Fehler!?\n";
    }

    return;
}


void write_rsc(std::ofstream& str, int min_len, int max_len)
// ==========================================================================================
{

 
	int my_min_len = min_len;
	int my_max_len = max_len;

   if (len<=0 || len<=max_len)
        return;

	if (max_len==0)
	{
		my_max_len = len;
		my_min_len = 0;
	}

	int my_len = my_max_len-my_min_len;
    str.write((char*)&my_len,sizeof(int));

    for (int i=my_min_len; i<my_max_len; i++)
    {
        str.write((char*)&(xx[i]),sizeof(double));
        str.write((char*)&(yy[i]),sizeof(double));
    }

}

// Downsaples the curve by taking only every 10th point.
bool write_rsc_ds(std::ofstream& str, const int min_len, const int max_len, const int samples)
// ==========================================================================================
{


	int my_min_len = min_len;
	int my_max_len = max_len;

    if (len<=0 || len<=max_len)
        return false;


    if (max_len!=0 && (len<=0 || len<=max_len))
        return;

    int new_len=0;
    double* new_xx = new double[len];
    double* new_yy = new double[len];


    for (int i=0; i<len; i++)
    {
        if (i%samples==0 || i==len)
        {
            new_xx[new_len] = xx[i];
            new_yy[new_len] = yy[i];
            new_len++;
        }
    }


	if (max_len!=0 && new_len<max_len)
		return;


	if (max_len==0)
	{
		my_max_len = new_len;
		my_min_len = 0;
	}

	int my_len = my_max_len-my_min_len;
    str.write((char*)&my_len,sizeof(int));

    for (i=my_min_len; i<my_max_len; i++)
    {
        str.write((char*)&(new_xx[i]),sizeof(double));
        str.write((char*)&(new_yy[i]),sizeof(double));
    }

    return true;

}

// Downsaples the curve by taking each point with a probability of 10% (or 20%).
bool write_rsc_rnd(std::ofstream& str, const int min_len, const int max_len, const int samples)
// ==========================================================================================
{

	int my_min_len = min_len;
	int my_max_len = max_len;

    if (max_len!=0 && (len<=0 || len<max_len))
        return;


    int new_len=0;
    double* new_xx = new double[len];
    double* new_yy = new double[len];


    for (int i=0; i<len; i++)
    {
        int r = rand();
        if (r%samples==0)
        {
            new_xx[new_len] = xx[i];
            new_yy[new_len] = yy[i];
            new_len++;
        }
    }


	if (max_len!=0 && new_len<max_len)
		return false;


	if (max_len==0)
	{
		my_max_len = new_len;
		my_min_len = 0;
	}

	int my_len = my_max_len-my_min_len;
    str.write((char*)&my_len,sizeof(int));

    for (i=my_min_len; i<my_max_len; i++)
    {
        str.write((char*)&(new_xx[i]),sizeof(double));
        str.write((char*)&(new_yy[i]),sizeof(double));
    }

    return true;

}


void get_noise_vector(double& rrx, double& rry,double s)
{
    
    // produce a noise vector of size <= 1.0
    double phi = (2.0*3.1415926535897932384626433832795*(double)rand())/(double)RAND_MAX;
    rrx = (s*(double)rand()*cos(phi))/(double)RAND_MAX;
    rry = (s*(double)rand()*sin(phi))/(double)RAND_MAX;

}

// Downsaples the curve by taking each point with a probability of 10% (or 20%)
// and adding some noise of up to 5 (or 10) pixels to each point.
bool write_rsc_rnd_noise(std::ofstream& str, const int min_len, const int max_len, const int samples, const int noise)
// ==========================================================================================
{
	
	int my_min_len = min_len;
	int my_max_len = max_len;

    if (len<=0 || len<max_len)
        return false;


    if (max_len!=0 && (len<=0 || len<max_len))
        return;

    int new_len=0;
    double* new_xx = new double[len];
    double* new_yy = new double[len];

    for (int i=0; i<len; i++)
    {
        int r = rand();
        if (r%samples==0)
        {
            
            new_xx[new_len] = xx[i]+rand()%noise;
            new_yy[new_len] = yy[i]+rand()%noise;
            new_len++;

        }
    }

	if (max_len!=0 && new_len<max_len)
		return false;

	if (max_len==0)
	{
		my_max_len = new_len;
		my_min_len = 0;
	}

	int my_len = my_max_len-my_min_len;
    str.write((char*)&my_len,sizeof(int));


    for (i=my_min_len; i<my_max_len; i++)
    {
        str.write((char*)&(new_xx[i]),sizeof(double));
        str.write((char*)&(new_yy[i]),sizeof(double));
    }

    return true;

}

// process squid-dataset
int main_squid(int argc, char** args)
{
 
    xx = NULL;
    yy = NULL;
    char filename[128];
    char basename[128];
    char outname[128];
    char numbuf[10];
    char numbuf2[10];

    len = 0;

    std::ifstream ifs;
    std::ofstream ofs;


	for (int curr_len=20; curr_len<200; curr_len+=10)
	{
        _itoa(curr_len,numbuf2,10);
		strcpy(basename,"w:\\data\\squid\\l");
		strcat(basename,numbuf2);


		for (int f_index=1; f_index<=1100; f_index++)
		{
			int fh;
            bool succ;

			_itoa(f_index,numbuf,10);
			strcpy(filename,"w:\\data\\squid\\kk");
			strcat(filename,numbuf);
			strcat(filename,".c");
			ifs.open(filename);
			read_squid(ifs);
			ifs.close();

			strcpy(outname,basename);
			strcat(outname,"\\kk");
			strcat(outname,numbuf);
			strcat(outname,".rsc");
			ofs.open(outname);
			write_rsc(ofs,0,0);
			ofs.close();

			strcpy(outname,basename);
			strcat(outname,"\\ds5_kk");
			strcat(outname,numbuf);
			strcat(outname,".rsc");
			ofs.open(outname);
			write_rsc_ds(ofs,0,0,5);
			ofs.close();


			strcpy(outname,basename);
			strcat(outname,"\\ds20_kk");
			strcat(outname,numbuf);
			strcat(outname,".rsc");
			ofs.open(outname);

			write_rsc_ds(ofs,0,0,20);
			ofs.close();


			/*strcpy(outname,basename);
			strcat(outname,"\\ds5\\qkk");
			strcat(outname,numbuf);
			strcat(outname,".rsc");
			ofs.open(outname);
			write_rsc_ds(ofs,curr_len/4,3*curr_len/4,5);
			ofs.close();*/


			/*strcpy(outname,basename);
			strcat(outname,"\\ds10\\kk");
			strcat(outname,numbuf);
			strcat(outname,".rsc");
			ofs.open(outname);
			if (len>=curr_len)
			{
				write_rsc_ds(ofs,0,curr_len,10);
			}
			ofs.close();

			strcpy(outname,basename);
			strcat(outname,"\\ds10\\qkk");
			strcat(outname,numbuf);
			strcat(outname,".rsc");
			ofs.open(outname);
			if (len>=curr_len)
			{
				write_rsc_ds(ofs,curr_len/4,3*curr_len/4,10);
			}
			ofs.close();

			strcpy(outname,basename);
			strcat(outname,"\\ds20\\kk");
			strcat(outname,numbuf);
			strcat(outname,".rsc");
			ofs.open(outname);
			if (len>=curr_len)
			{
				write_rsc_ds(ofs,0,curr_len,20);
			}
			ofs.close();

			strcpy(outname,basename);
			strcat(outname,"\\ds20\\qkk");
			strcat(outname,numbuf);
			strcat(outname,".rsc");
			ofs.open(outname);
			if (len>=curr_len)
			{
				write_rsc_ds(ofs,curr_len/4,3*curr_len/4,20);
			}
			ofs.close();*/


			strcpy(outname,basename);
			strcat(outname,"\\rnd5_kk");
			strcat(outname,numbuf);
			strcat(outname,".rsc");
			ofs.open(outname);
			if (len>=curr_len)
			{
				write_rsc_rnd_noise(ofs,0,0,5,5);
			}
			ofs.close();

			strcpy(outname,basename);
			strcat(outname,"\\rnd20_kk");
			strcat(outname,numbuf);
			strcat(outname,".rsc");
            ofs.open(outname,std::ios::binary);
			if (len>=curr_len)
			{

				succ = write_rsc_rnd_noise(ofs,curr_len/4,3*curr_len/4,5,5);

			}
			ofs.close();
            if (!succ)
                remove(outname);



		}
	}

    return 0;

}

// process sattelite-dataset
int main_sat(int argc, char** args)
{

    xx = NULL;
    yy = NULL;
    char filename[128];
    char basename[128];
    char outname[128];
    char numbuf[10];
    char numbuf2[10];

    len = 0;

    std::ifstream ifs;
    std::ofstream ofs;


	strcpy(filename,"w:\\data\\sattelite\\9-1.rsc");
    ifs.open(filename, std::ios::binary);
	read_rsc(ifs);
	ifs.close();

    for (int i=1; i<=10; i++)
    {
        strcpy(outname,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(outname,numbuf);
        strcat(outname,"-9-1.rsc");
        ofs.open(outname, std::ios::binary);
	    write_rsc_pattern(ofs,i*1000);
	    ofs.close();
    }

    strcpy(filename,"w:\\data\\sattelite\\9-2.rsc");
    ifs.open(filename, std::ios::binary);
	read_rsc(ifs);
	ifs.close();

    for (i=1; i<=10; i++)
    {
        strcpy(outname,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(outname,numbuf);
        strcat(outname,"-9-2.rsc");
        ofs.open(outname, std::ios::binary);
	    write_rsc_pattern(ofs,i*1000);
	    ofs.close();
    }

	strcpy(filename,"w:\\data\\sattelite\\9-3.rsc");
    ifs.open(filename, std::ios::binary);
	read_rsc(ifs);
	ifs.close();

    for (i=1; i<=10; i++)
    {
        strcpy(outname,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(outname,numbuf);
        strcat(outname,"-9-3.rsc");
        ofs.open(outname, std::ios::binary);
	    write_rsc_pattern(ofs,i*1000);
	    ofs.close();
    }

	strcpy(filename,"w:\\data\\sattelite\\9-4.rsc");
    ifs.open(filename, std::ios::binary);
	read_rsc(ifs);
	ifs.close();

    for (i=1; i<=10; i++)
    {
        strcpy(outname,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(outname,numbuf);
        strcat(outname,"-9-4.rsc");
        ofs.open(outname, std::ios::binary);
	    write_rsc_pattern(ofs,i*1000);
	    ofs.close();
    }

	strcpy(filename,"w:\\data\\sattelite\\9-5.rsc");
    ifs.open(filename, std::ios::binary);
	read_rsc(ifs);
	ifs.close();

    for (i=1; i<=10; i++)
    {
        strcpy(outname,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(outname,numbuf);
        strcat(outname,"-9-5.rsc");
        ofs.open(outname, std::ios::binary);
	    write_rsc_pattern(ofs,i*1000);
	    ofs.close();
    }

	strcpy(filename,"w:\\data\\sattelite\\9-6.rsc");
    ifs.open(filename, std::ios::binary);
	read_rsc(ifs);
	ifs.close();

    for (i=1; i<=10; i++)
    {
        strcpy(outname,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(outname,numbuf);
        strcat(outname,"-9-6.rsc");
        ofs.open(outname, std::ios::binary);
	    write_rsc_pattern(ofs,i*1000);
	    ofs.close();
    }

}
   

int get_rnd(int imin, int imax)
{
    return imin + ((imax-imin)*rand())/(RAND_MAX+1);
}

// process sattelite-dataset
int main_sat_query(int argc, char** args)
{

    xx = NULL;
    yy = NULL;
    char filename[128];
    char basename[128];
    char outname[128];
    char numbuf[10];
    char numbuf2[10];

    len = 0;

    std::ifstream ifs;
    std::ofstream ofs;

    for (int i=1; i<=10; i++)
    {
        
        strcpy(filename,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(filename,numbuf);
        strcpy(outname,filename);
        strcpy(basename,filename);
        basename[0]='q';
        strcat(filename,"-9-1.rsc");
        
        ifs.open(filename, std::ios::binary);
    	read_rsc(ifs);
    	ifs.close();

        for (int j=0; j<20; j++)
        {
            char numbuf2[5];
            itoa(j,numbuf2,10);
            strcpy(outname,basename);
            strcat(outname,"-9-1-");
            strcat(outname,numbuf2);
            strcat(outname,".rsc");
            ofs.open(outname, std::ios::binary);
            int l_x = get_rnd(bb_min_x,bb_min_x+(bb_max_x-bb_min_x)/2);
            int l_y = get_rnd(bb_min_y,bb_min_y+(bb_max_y-bb_min_y)/2);
            int u_x = l_x+(j+1)*(bb_max_x-bb_min_x)/40;
            int u_y = l_y+(j+1)*(bb_max_y-bb_min_y)/40;
            int h;
            if (l_x>u_x) {h=l_x; l_x=u_x; u_x=h;}
            if (l_y>u_y) {h=l_y; l_y=u_y; u_y=h;}
	        write_rsc_query(ofs,l_x,l_y,u_x,u_y);
	        ofs.close();
        }

    }

    for (i=1; i<=10; i++)
    {
        
        strcpy(filename,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(filename,numbuf);
        strcpy(outname,filename);
        strcpy(basename,filename);
        basename[0]='q';
        strcat(filename,"-9-2.rsc");
        
        ifs.open(filename, std::ios::binary);
    	read_rsc(ifs);
    	ifs.close();

        for (int j=0; j<20; j++)
        {
            char numbuf2[5];
            itoa(j,numbuf2,10);
            strcpy(outname,basename);
            strcat(outname,"-9-2-");
            strcat(outname,numbuf2);
            strcat(outname,".rsc");
            ofs.open(outname, std::ios::binary);
            int l_x = get_rnd(bb_min_x,bb_min_x+(bb_max_x-bb_min_x)/2);
            int l_y = get_rnd(bb_min_y,bb_min_y+(bb_max_y-bb_min_y)/2);
            int u_x = l_x+(j+1)*(bb_max_x-bb_min_x)/40;
            int u_y = l_y+(j+1)*(bb_max_y-bb_min_y)/40;
            int h;
            if (l_x>u_x) {h=l_x; l_x=u_x; u_x=h;}
            if (l_y>u_y) {h=l_y; l_y=u_y; u_y=h;}
	        write_rsc_query(ofs,l_x,l_y,u_x,u_y);
	        ofs.close();
        }
    }

    for (i=1; i<=10; i++)
    {
        
        strcpy(filename,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(filename,numbuf);
        strcpy(outname,filename);
        strcpy(basename,filename);
        basename[0]='q';
        strcat(filename,"-9-3.rsc");
        
        ifs.open(filename, std::ios::binary);
    	read_rsc(ifs);
    	ifs.close();

        for (int j=0; j<20; j++)
        {
            char numbuf2[5];
            itoa(j,numbuf2,10);
            strcpy(outname,basename);
            strcat(outname,"-9-3-");
            strcat(outname,numbuf2);
            strcat(outname,".rsc");
            ofs.open(outname, std::ios::binary);
            int l_x = get_rnd(bb_min_x,bb_min_x+(bb_max_x-bb_min_x)/2);
            int l_y = get_rnd(bb_min_y,bb_min_y+(bb_max_y-bb_min_y)/2);
            int u_x = l_x+(j+1)*(bb_max_x-bb_min_x)/40;
            int u_y = l_y+(j+1)*(bb_max_y-bb_min_y)/40;
            int h;
            if (l_x>u_x) {h=l_x; l_x=u_x; u_x=h;}
            if (l_y>u_y) {h=l_y; l_y=u_y; u_y=h;}
	        write_rsc_query(ofs,l_x,l_y,u_x,u_y);
	        ofs.close();
        }

    }

    for (i=1; i<=10; i++)
    {
        
        strcpy(filename,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(filename,numbuf);
        strcpy(outname,filename);
        strcpy(basename,filename);
        basename[0]='q';
        strcat(filename,"-9-4.rsc");
        
        ifs.open(filename, std::ios::binary);
    	read_rsc(ifs);
    	ifs.close();

        for (int j=0; j<20; j++)
        {
            char numbuf2[5];
            itoa(j,numbuf2,10);
            strcpy(outname,basename);
            strcat(outname,"-9-4-");
            strcat(outname,numbuf2);
            strcat(outname,".rsc");
            ofs.open(outname, std::ios::binary);
            int l_x = get_rnd(bb_min_x,bb_min_x+(bb_max_x-bb_min_x)/2);
            int l_y = get_rnd(bb_min_y,bb_min_y+(bb_max_y-bb_min_y)/2);
            int u_x = l_x+(j+1)*(bb_max_x-bb_min_x)/40;
            int u_y = l_y+(j+1)*(bb_max_y-bb_min_y)/40;
            int h;
            if (l_x>u_x) {h=l_x; l_x=u_x; u_x=h;}
            if (l_y>u_y) {h=l_y; l_y=u_y; u_y=h;}
	        write_rsc_query(ofs,l_x,l_y,u_x,u_y);
	        ofs.close();
        }

    }

    for (i=1; i<=10; i++)
    {
        
        strcpy(filename,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(filename,numbuf);
        strcpy(outname,filename);
        strcpy(basename,filename);
        basename[0]='q';
        strcat(filename,"-9-5.rsc");
        
        ifs.open(filename, std::ios::binary);
    	read_rsc(ifs);
    	ifs.close();

        for (int j=0; j<20; j++)
        {
            char numbuf2[5];
            itoa(j,numbuf2,10);
            strcpy(outname,basename);
            strcat(outname,"-9-5-");
            strcat(outname,numbuf2);
            strcat(outname,".rsc");
            ofs.open(outname, std::ios::binary);
            int l_x = get_rnd(bb_min_x,bb_min_x+(bb_max_x-bb_min_x)/2);
            int l_y = get_rnd(bb_min_y,bb_min_y+(bb_max_y-bb_min_y)/2);
            int u_x = l_x+(j+1)*(bb_max_x-bb_min_x)/40;
            int u_y = l_y+(j+1)*(bb_max_y-bb_min_y)/40;
            int h;
            if (l_x>u_x) {h=l_x; l_x=u_x; u_x=h;}
            if (l_y>u_y) {h=l_y; l_y=u_y; u_y=h;}
	        write_rsc_query(ofs,l_x,l_y,u_x,u_y);
	        ofs.close();
        }

    }

    for (i=1; i<=10; i++)
    {
        
        strcpy(filename,"p");
        char numbuf[10];
        itoa(i*1000,numbuf,10);
        strcat(filename,numbuf);
        strcpy(outname,filename);
        strcpy(basename,filename);
        basename[0]='q';
        strcat(filename,"-9-6.rsc");
        
        ifs.open(filename, std::ios::binary);
    	read_rsc(ifs);
    	ifs.close();

        for (int j=0; j<20; j++)
        {
            char numbuf2[5];
            itoa(j,numbuf2,10);
            strcpy(outname,basename);
            strcat(outname,"-9-6-");
            strcat(outname,numbuf2);
            strcat(outname,".rsc");
            ofs.open(outname, std::ios::binary);
            int l_x = get_rnd(bb_min_x,bb_min_x+(bb_max_x-bb_min_x)/2);
            int l_y = get_rnd(bb_min_y,bb_min_y+(bb_max_y-bb_min_y)/2);
            int u_x = l_x+(j+1)*(bb_max_x-bb_min_x)/40;
            int u_y = l_y+(j+1)*(bb_max_y-bb_min_y)/40;
            int h;
            if (l_x>u_x) {h=l_x; l_x=u_x; u_x=h;}
            if (l_y>u_y) {h=l_y; l_y=u_y; u_y=h;}
	        write_rsc_query(ofs,l_x,l_y,u_x,u_y);
	        ofs.close();
        }

    }

}
   
int main(int argc, char** args)
{
    //main_squid(argc,args);
    main_sat_query(argc,args);
    //main_sat(argc,args);
}

